import json
import logging.config
import os
from configparser import ConfigParser
from pathlib import Path
import requests
from google.cloud import kms, storage
import boto3
import base64
from cryptography.fernet import Fernet
from infoworks.core.iw_authentication import is_token_valid, get_bearer_token, aes_encrypt
from infoworks.sdk import local_configurations
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry

from infoworks.sdk.utils import IWUtils


class TimeoutHTTPAdapter(HTTPAdapter):
    def __init__(self, *args, **kwargs):
        self.timeout = local_configurations.REQUEST_TIMEOUT_IN_SEC
        if "timeout" in kwargs:
            self.timeout = kwargs["timeout"]
            del kwargs["timeout"]
        super().__init__(*args, **kwargs)

    def send(self, request, **kwargs):
        timeout = kwargs.get("timeout")
        if timeout is None:
            kwargs["timeout"] = self.timeout
        return super().send(request, **kwargs)


def initialise_http_client():
    retries = Retry(total=local_configurations.MAX_RETRIES, backoff_factor=1,
                    status_forcelist=[429, 500, 502, 503, 504])
    http = requests.Session()
    adapter = TimeoutHTTPAdapter(max_retries=retries)
    http.mount("https://", adapter)
    http.mount("http://", adapter)
    return http


class BaseClient(object):
    def __init__(self):
        # print("Inside BaseClient")
        self.client_config = {
            'protocol': None,
            'ip': None,
            'port': None,
            'refresh_token': None,
            'bearer_token': None,
            'default_environment_id': None,
            'default_storage_id': None,
            'default_compute_id': None
        }
        self.secrets_config = {"custom_secrets_read": False}
        self.mappings = {}
        self.http = initialise_http_client()
        log_path = Path(local_configurations.LOG_LOCATION)
        if os.path.isdir(log_path.parent.absolute()):
            logging.basicConfig(filename=local_configurations.LOG_LOCATION, filemode='w',
                                format='%(asctime)s - %(module)s - %(pathname)s - %(lineno)d - %(levelname)s - %(message)s',
                                level=logging.DEBUG, datefmt='%Y-%m-%d %H:%M:%S')
        else:
            logging.basicConfig(
                format='%(asctime)s - %(module)s - %(pathname)s - %(lineno)d - %(levelname)s - %(message)s',
                level=logging.DEBUG, datefmt='%Y-%m-%d %H:%M:%S')
        self.logger = logging.getLogger('infoworks_sdk_logs')

    def initialize_client(self, protocol=None, ip=None, port=None, ):
        """
        initialize the client
        :param protocol: protocol to be used for server call
        :type protocol: String
        :param ip: Client IP
        :type ip: String
        :param port: Client Port
        :type port: String
        """

        if all(v is not None for v in [protocol, ip, port]):
            self.client_config['protocol'] = protocol
            self.client_config['ip'] = ip
            self.client_config['port'] = port
        else:
            logging.error(ValueError("Pass valid values"))
            raise ValueError("Pass valid values")

    def initialize_client_with_defaults(self, protocol, ip, port, refresh_token, default_environment_id=None,
                                        default_storage_id=None, default_compute_id=None):
        """
        initializes the client and a user with given configuration
        :param protocol: protocol to be used for server call
        :type protocol: String
        :param ip: ip address of the server
        :type ip: String
        :param port: port on which the rest service resides
        :type port: String
        :param refresh_token: refresh_token of the user
        :type refresh_token: String
        :param default_compute_id: Pass the default compute id to be used for all the artifacts to be created
        :param default_storage_id: Pass the default storage id to be used for all the artifacts to be created
        :param default_environment_id: Pass the default environment id to be used for all the artifacts to be created
        """
        self.client_config['refresh_token'] = refresh_token
        self.client_config['bearer_token'] = get_bearer_token(protocol, ip, port, refresh_token)
        self.initialize_client(protocol, ip, port)
        self.client_config[
            'default_environment_id'] = default_environment_id if default_environment_id is not None else None
        self.client_config['default_storage_id'] = default_storage_id if default_storage_id is not None else None
        self.client_config['default_compute_id'] = default_compute_id if default_compute_id is not None else None

    def get_configurations(self):
        """
        returns client configurations
        :return: configuration dictionary
        """
        return self.client_config

    def regenerate_bearer_token_if_needed(self, headers):
        if not is_token_valid(self.client_config, self.http):
            self.client_config['bearer_token'] = get_bearer_token(protocol=self.client_config["protocol"],
                                                                  ip=self.client_config["ip"],
                                                                  port=self.client_config["port"],
                                                                  refresh_token=self.client_config["refresh_token"])
            headers = IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
        return headers

    def call_api(self, method, url, headers=None, data=None):
        # headers = self.regenerate_bearer_token_if_needed(headers)
        if method.upper() == "GET":
            self.logger.info(f"Calling {url}")
            response = self.http.get(url, headers=headers, timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC,
                                     verify=False, data=data)
            if response.status_code == 406:
                headers = self.regenerate_bearer_token_if_needed(headers)
                return self.http.get(url, headers=headers, timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC,
                                     verify=False, data=data)
            else:
                return response
        elif method.upper() == "POST":
            self.logger.info(f"Calling {url}")
            response = self.http.post(url, headers=headers, json=data,
                                      timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC, verify=False)
            if response.status_code == 406:
                headers = self.regenerate_bearer_token_if_needed(headers)
                return self.http.post(url, headers=headers, json=data,
                                      timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC, verify=False)
            else:
                return response
        elif method.upper() == "PUT":
            self.logger.info(f"Calling {url}")
            response = self.http.put(url, headers=headers, json=data,
                                     timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC, verify=False)
            if response.status_code == 406:
                headers = self.regenerate_bearer_token_if_needed(headers)
                return self.http.put(url, headers=headers, json=data,
                                     timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC, verify=False)
            else:
                return response
        elif method.upper() == "PATCH":
            self.logger.info(f"Calling {url}")
            response = self.http.patch(url, headers=headers, json=data,
                                       timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC, verify=False)
            if response.status_code == 406:
                headers = self.regenerate_bearer_token_if_needed(headers)
                return self.http.put(url, headers=headers, json=data,
                                     timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC, verify=False)
            else:
                return response
        elif method.upper() == "DELETE":
            self.logger.info(f"Calling {url}")
            response = self.http.delete(url, headers=headers, timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC,
                                        verify=False)
            if response.status_code == 406:
                headers = self.regenerate_bearer_token_if_needed(headers)
                return self.http.delete(url, headers=headers, timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC,
                                        verify=False)
            else:
                return response

    def get_mappings_from_config_file(self, ini_config_file_path):
        config = ConfigParser()
        config.optionxform = str
        config.read(ini_config_file_path)
        if "environment" in config:
            default_environment_id = config.get("environment", "environment_id")
            default_storage_id = config.get("environment", "storage_id")
            default_interactive_id = config.get("environment", "interactive_id")
            self.client_config["default_environment_id"] = default_environment_id
            self.client_config["default_storage_id"] = default_storage_id
            self.client_config["default_compute_id"] = default_interactive_id
        if "environment_mappings" in config:
            environment_mappings = dict(config.items("environment_mappings"))
            self.mappings["environment_mappings"] = environment_mappings
        if "storage_mappings" in config:
            storage_mappings = dict(config.items("storage_mappings"))
            self.mappings["storage_mappings"] = storage_mappings
        if "compute_mappings" in config:
            compute_mappings = dict(config.items("compute_mappings"))
            self.mappings["compute_mappings"] = compute_mappings
        if "gcp_details" in config:
            gcp_details = dict(config.items("gcp_details"))
            self.mappings["gcp_details"] = gcp_details
        if "gcp_project_id_mappings" in config:
            gcp_project_id_mappings = dict(config.items("gcp_project_id_mappings"))
            self.mappings["gcp_project_id_mappings"] = gcp_project_id_mappings
        if "service_json_mappings" in config:
            service_json_mappings = dict(config.items("service_json_mappings"))
            self.mappings["service_json_mappings"] = service_json_mappings
        if "table_group_compute_mappings" in config:
            table_group_compute_mappings = dict(config.items("table_group_compute_mappings"))
            self.mappings["table_group_compute_mappings"] = table_group_compute_mappings
        if "kms" in config:
            kms = dict(config.items("kms"))
            self.mappings["kms"] = kms

    def _topological_sort_grouping(self, g):
        # copy the graph
        _g = g.copy()
        res = []
        # while _g is not empty
        while _g:
            zero_indegree = [v for v, d in _g.in_degree() if d == 0]
            res.append(zero_indegree)
            _g.remove_nodes_from(zero_indegree)
        return res

    def get_all_secrets(self, secret_type, ini_config_file_path=None, keys=None):
        # Read all secrets from file and load inmemory
        # Encrypt using infoworks encoding
        if secret_type == "gcp_kms" and ini_config_file_path is not None:
            config = ConfigParser()
            config.optionxform = str
            config.read(ini_config_file_path)
            # Cloud KMS CryptoKey Encrypter
            service_json_path = config.get("gcp_kms", "service_json_path")
            # Reading Config Parameters
            projectId = config.get("gcp_kms", "projectId")
            locationId = config.get("gcp_kms", "locationId")
            keyRingId = config.get("gcp_kms", "keyRingId")
            keyId = config.get("gcp_kms", "keyId")
            bucketName = config.get("gcp_kms", "bucketName")
            blobName = config.get("gcp_kms", "encrypted_key_filename")
            # Reading Encrypted Password File from Cloud Storage
            storageClient = storage.Client.from_service_account_json(service_json_path)
            bucket = storageClient.get_bucket(bucketName)
            blobFile = bucket.get_blob(blobName)
            cipherText = blobFile.download_as_string()
            kmsClient = kms.KeyManagementServiceClient.from_service_account_json(service_json_path)
            keyName = kmsClient.crypto_key_path(projectId, locationId, keyRingId, keyId)
            # Decrypt Password
            try:
                decryptPassword = kmsClient.decrypt(request={'name': keyName, 'ciphertext': cipherText})
                plainText = str(decryptPassword.plaintext, 'utf-8')
                plainTextStr = plainText.strip()
                passwordsDictPlain = json.loads(plainTextStr)
                passwordsDictEncrypted = {}
                for key in passwordsDictPlain:
                    passwordsDictEncrypted[key] = aes_encrypt(passwordsDictPlain[key])
                self.secrets_config = passwordsDictEncrypted
            except Exception as e:
                print(str(e))
                return ""
        elif secret_type == "aws_kms" and ini_config_file_path is not None:
            config = ConfigParser()
            config.optionxform = str
            config.read(ini_config_file_path)
            # Read the encrypted file into memory
            encrypted_filename = config.get("aws_kms", "encrypted_filename")
            available_keys = dict(config.items("aws_kms")).keys()
            if "access_id" in available_keys and "secret_key" in available_keys:
                ACCESS_KEY = config.get("aws_kms", "access_id")
                SECRET_KEY = config.get("aws_kms", "secret_key")
            else:
                ACCESS_KEY, SECRET_KEY = None, None
            # Read
            NUM_BYTES_FOR_LEN = 4
            try:
                with open(encrypted_filename, 'rb') as file:
                    file_contents = file.read()
            except IOError as e:
                self.logger.error(str(e))
                return False
            data_key_encrypted_len = int.from_bytes(file_contents[:NUM_BYTES_FOR_LEN],
                                                    byteorder='big') + NUM_BYTES_FOR_LEN
            data_key_encrypted = file_contents[NUM_BYTES_FOR_LEN:data_key_encrypted_len]
            # Decrypt the data key before using it
            if ACCESS_KEY is not None and SECRET_KEY is not None:
                kms_client = boto3.client('kms', aws_access_key_id=ACCESS_KEY, aws_secret_access_key=SECRET_KEY)
            else:
                kms_client = boto3.client('kms')
            try:
                response = kms_client.decrypt(CiphertextBlob=data_key_encrypted)
                data_key_plaintext = base64.b64encode((response['Plaintext']))
            except Exception as e:
                self.logger.error(str(e))
                return False

            if data_key_plaintext is None:
                return False
            # Decrypt the rest of the file
            f = Fernet(data_key_plaintext)
            file_contents_decrypted = f.decrypt(file_contents[data_key_encrypted_len:])
            passwordsDictPlain = json.loads(file_contents_decrypted)
            passwordsDictEncrypted = {}
            for key in passwordsDictPlain:
                passwordsDictEncrypted[key] = aes_encrypt(passwordsDictPlain[key])
            self.secrets_config = passwordsDictEncrypted
            return True
        elif secret_type == "aws_secrets":
            ACCESS_KEY, SECRET_KEY = None, None
            region = 'us-west-2'
            if ini_config_file_path is not None:
                config = ConfigParser()
                config.optionxform = str
                config.read(ini_config_file_path)
                available_keys = dict(config.items("aws_secrets")).keys()
                if "access_id" in available_keys and "secret_key" in available_keys:
                    ACCESS_KEY = config.get("aws_secrets", "access_id")
                    SECRET_KEY = config.get("aws_secrets", "secret_key")
                else:
                    ACCESS_KEY, SECRET_KEY = None, None
                if "region" in available_keys:
                    region = config.get("aws_secrets", "region")
            if keys:
                output_values = []
                for key in keys.split(","):
                    secrets_client = boto3.client("secretsmanager", region_name=region)
                    kwargs = {'SecretId': key}
                    try:
                        value = secrets_client.get_secret_value(**kwargs)["SecretString"]
                        self.secrets_config[key] = value
                        output_values.append(value)
                    except Exception as e:
                        self.logger.error(str(e))
                return output_values
            else:
                # Get all the secrets
                if ACCESS_KEY is not None and SECRET_KEY is not None:
                    secrets_client = boto3.client("secretsmanager", aws_access_key_id=ACCESS_KEY,
                                                  aws_secret_access_key=SECRET_KEY, region=region)
                else:
                    secrets_client = boto3.client("secretsmanager",region=region)
                temp = secrets_client.list_secrets()["SecretList"]
                all_secrets = [i["Name"] for i in temp]
                for key in all_secrets:
                    kwargs = {'SecretId': key}
                    try:
                        value = secrets_client.get_secret_value(**kwargs)["SecretString"]
                        self.secrets_config[key] = value
                    except Exception as e:
                        self.logger.error(str(e))
        elif secret_type == "azure_keyvault" and ini_config_file_path is not None:
            from azure.keyvault.secrets import SecretClient
            from azure.identity import ClientSecretCredential
            config = ConfigParser()
            config.optionxform = str
            config.read(ini_config_file_path)
            # Read the encrypted file into memory
            available_keys = dict(config.items("azure_keyvault")).keys()
            if all(x in ['keyVaultName', 'tenant_id', 'client_id', 'client_secret'] for x in available_keys):
                keyVaultName = config.get("azure_keyvault", "keyVaultName")
                KVUri = f"https://{keyVaultName}.vault.azure.net"
                tenant_id = config.get("azure_keyvault", "tenant_id")
                client_id = config.get("azure_keyvault", "client_id")
                client_secret = config.get("azure_keyvault", "client_secret")
                credential = ClientSecretCredential(tenant_id, client_id, client_secret)
                client = SecretClient(vault_url=KVUri, credential=credential)
                output_values=[]
                if keys:
                    for key in keys.split(","):
                        retrieved_secret = client.get_secret(key)
                        self.secrets_config[key] = aes_encrypt(retrieved_secret.value)
                        output_values.append(aes_encrypt(retrieved_secret.value))
                else:
                    secret_properties = client.list_properties_of_secrets()
                    all_secrets = [i.name for i in secret_properties]
                    for key in all_secrets:
                        retrieved_secret = client.get_secret(key)
                        self.secrets_config[key] = aes_encrypt(retrieved_secret.value)
                        output_values.append(retrieved_secret.value)
                return output_values

        self.secrets_config["custom_secrets_read"] = True



#!/usr/bin/env python
import traceback

import requests
import logging
import infoworks.error
from infoworks.sdk import url_builder, local_configurations
from infoworks.sdk.utils import IWUtils
from Cryptodome.Cipher import AES
import base64
from Cryptodome.Random import get_random_bytes
from Cryptodome.Protocol.KDF import PBKDF2
from Cryptodome.Hash import SHA256


def validate_bearer_token(protocol, ip, port, delegation_token):
    client_config = {"protocol": protocol, "ip": ip, "port": port}
    headers = {'Authorization': 'Bearer ' + delegation_token, 'Content-Type': 'application/json'}
    response = IWUtils.ejson_deserialize(
        requests.get(url_builder.validate_bearer_token_url(client_config), headers=headers, verify=False).content)
    if len(response.get("result", {})) > 0:
        return response.get("result")["is_valid"]
    else:
        return False


def get_bearer_token(protocol, ip, port, refresh_token, http_client=None):
    if any([protocol is None, ip is None, port is None, refresh_token is None]):
        raise infoworks.error.AuthenticationError("Some of the arguments are invalid or has None")
    client_config = {"protocol": protocol, "ip": ip, "port": port}
    headers = {'Authorization': 'Basic ' + refresh_token, 'Content-Type': 'application/json'}
    if http_client is not None:
        response = http_client.get(url_builder.get_bearer_token_url(client_config), headers=headers,
                                   timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC,
                                   verify=False)
    else:
        response = requests.get(url_builder.get_bearer_token_url(client_config), headers=headers,
                                timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC,
                                verify=False)
    if response:
        delegation_token = response.json().get("result").get("authentication_token")
        return delegation_token
    else:
        traceback.print_stack()
        raise infoworks.error.AuthenticationError("Refresh token invalid")


def is_token_valid(client_config, http_client=None):
    try:
        if client_config['bearer_token'] is None:
            logging.info('Invalid Delegation token: ' + str(client_config['bearer_token']))
            return False
        if http_client is not None:
            result = http_client.get(
                url_builder.validate_bearer_token_url(client_config),
                timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC,
                headers=IWUtils.get_default_header_for_v3(client_config['bearer_token']), verify=False
            )
        else:
            result = requests.get(
                url_builder.validate_bearer_token_url(client_config),
                timeout=local_configurations.REQUEST_TIMEOUT_IN_SEC,
                headers=IWUtils.get_default_header_for_v3(client_config['bearer_token']), verify=False
            )
        response = IWUtils.ejson_deserialize(result.content)
        logging.info('Response from rest api: ' + str(response))
        return response.get('result', {}).get('is_valid', False)
    except Exception as e:
        logging.error('Failed to check token validity: ' + str(e))
        return False


def aes_encrypt(data):
    salt = get_random_bytes(16)
    iv = get_random_bytes(12)
    key = PBKDF2("infoworks", salt, dkLen=256, count=65536, hmac_hash_module=SHA256)
    key = key[0:32]
    cipher = AES.new(key, AES.MODE_GCM, nonce=iv)
    ciphertext, auth_tag = cipher.encrypt_and_digest(bytes(data, 'utf-8'))
    return base64.b64encode(salt + iv + ciphertext + auth_tag).decode('utf-8')


def aes_decrypt(encrypted_data):
    ciphertext = base64.b64decode(encrypted_data)
    salt = ciphertext[0:16]
    iv = ciphertext[16:28]
    auth_tag = ciphertext[-16:]
    text = ciphertext[28:-16]
    key = PBKDF2("infoworks", salt, dkLen=256, count=65536, hmac_hash_module=SHA256)
    key = key[0:32]
    cipher = AES.new(key, AES.MODE_GCM, nonce=iv)
    decrypt_text = cipher.decrypt_and_verify(text, auth_tag)
    return decrypt_text.decode('utf-8')



import time
import traceback

from infoworks.error import SourceError
from infoworks.sdk import url_builder, local_configurations
from infoworks.sdk.base_client import BaseClient
from infoworks.sdk.local_configurations import Response, ErrorCode, SourceMappings
from infoworks.sdk.source_response import SourceResponse
from infoworks.sdk.utils import IWUtils


class SourceClient(BaseClient):

    def __init__(self):
        super(SourceClient, self).__init__()

    def poll_job(self, source_id=None, job_id=None, poll_timeout=local_configurations.POLLING_TIMEOUT,
                 polling_frequency=local_configurations.POLLING_FREQUENCY_IN_SEC,
                 retries=local_configurations.NUM_POLLING_RETRIES):
        """
                Polls Infoworks source job and returns its status
                :param source_id: Source Identifier
                :type source_id: String
                :param job_id: Job Identifier
                :type job_id: String
                :param poll_timeout: Polling timeout(default : 1200 seconds)
                :type poll_timeout: Integer
                :param polling_frequency: Polling Frequency(default : 15 seconds)
                :type polling_frequency: Integer
                :param retries: Number of Retries (default : 3)
                :type retries: Integer
                :return: response Object with Job status
        """
        failed_count = 0
        response = {}
        timeout = time.time() + poll_timeout
        while True:
            if time.time() > timeout:
                break
            try:
                self.logger.info(f"Failed poll job status count: {failed_count}")
                job_monitor_url = url_builder.get_job_status_url(self.client_config, job_id)
                response = IWUtils.ejson_deserialize(self.call_api("GET", job_monitor_url,
                                                                   IWUtils.get_default_header_for_v3(
                                                                       self.client_config['bearer_token'])).content)
                result = response.get('result', {})
                if len(result) != 0:
                    job_status = result["status"]
                    self.logger.info(
                        "Job poll status : " + result["status"] + "Job completion percentage: " + str(result.get(
                            "percentage", 0)))
                    if job_status in ["completed", "failed", "aborted"]:
                        if job_status == "completed":
                            return SourceResponse.parse_result(status=Response.Status.SUCCESS, source_id=source_id,
                                                               job_id=job_id)
                        else:
                            return SourceResponse.parse_result(status=job_status, source_id=source_id,
                                                               job_id=job_id,response=response)
                else:
                    self.logger.error(f"Error occurred during job {job_id} status poll")
                    if failed_count >= retries - 1:
                        return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                           error_code=ErrorCode.GENERIC_ERROR,
                                                           error_desc=response, job_id=job_id,
                                                           source_id=source_id)
                    failed_count = failed_count + 1
            except Exception as e:
                self.logger.exception("Error occurred during job status poll")
                self.logger.info(str(e))
                if failed_count >= retries - 1:
                    # traceback.print_stack()
                    print(response)
                    raise SourceError(response.get("message", "Error occurred during job status poll"))
                failed_count = failed_count + 1
            time.sleep(polling_frequency)

    def create_source(self, source_config=None):
        """
        Create a new Infoworks source
        :param source_config: a JSON object containing source configurations
        :type source_config: JSON Object

        source_config_example = {
            "name": "source_name",
            "type": "rdbms",
            "sub_type": "subtype",
            "data_lake_path": "datalake_path",
            "environment_id": "environment_id",
            "storage_id": "storage_id",
            "is_source_ingested": True
        }
        :return: response dict
        """
        response = None
        try:
            if source_config is None:
                self.logger.error('Invalid source configuration. Cannot create a new source.')
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_desc='Invalid source configuration. '
                                                              'Cannot create a new source.')
            response = IWUtils.ejson_deserialize(
                self.call_api("POST", url_builder.create_source_url(self.client_config),
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              source_config).content)

            result = response.get('result', {})
            source_id = result.get('id', None)

            if source_id is None:
                self.logger.error('Source failed to create.')
                return SourceResponse.parse_result(status=Response.Status.FAILED, error_desc=response)

            source_id = str(source_id)
            self.logger.info('Source {id} has been created.'.format(id=source_id))
            return SourceResponse.parse_result(status=Response.Status.SUCCESS, source_id=source_id,response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to create a new source.')
            print('Error occurred while trying to create a new source.')
            raise SourceError(response.get("message", "Error occurred while trying to create a new source."))

    def configure_source_connection(self, source_id, connection_object, read_passwords_from_secrets=False):
        """
        Function to configure the source connection
        :param read_passwords_from_secrets: True/False. If True then the passwords are read from the secret manager info provided
        :param source_id: source identifier entity id
        :type source_id: String
        :param connection_object: The json dict containing the source connection details
        :type connection_object: Json object

        "connection_obj_rdbms": {
                "enable_log_based_cdc": false,
                "driver_name": "oracle.jdbc.driver.OracleDriver",
                "driver_version": "v2",
                "connection_url": "jdbc:oracle:thin:@52.73.246.109:1521:xe",
                "username": "automation_db",
                "password": "eEBcRuPkw0zh9oIPvKnal+1BNKmFH5TfdI1ieDinruUv47Z5+f/oPjb+uyqUmfcQusM2DjoHc3OM", # can be iwx encrypted password or plain text password
                "connection_mode": "JDBC",
                "database": "ORACLE"
                }

        "connection_obj_csv" = {
                "source_base_path_relative": "",
                "source_base_path": "",
                "storage": {
                    "storage_type": "",
                    "cloud_type": "",
                    "access_id": "",
                    "secret_key": "",
                    "account_type": "",
                    "access_type": ""
                }
            }
        :return: response dict
        """

        if read_passwords_from_secrets:
            src_name = self.get_sourcename_from_id(source_id)
            if src_name is not None:
                encrypted_key_name = "iwxsrc-" + src_name + "-password"
                decrypt_pass = self.secrets_config.get(encrypted_key_name, "")
                connection_object['password'] = decrypt_pass if decrypt_pass else connection_object.get('password', "")
        try:
            source_connection_configuration_url = url_builder.get_source_connection_details_url(self.client_config,
                                                                                                source_id)
            response = IWUtils.ejson_deserialize(self.call_api("PUT", source_connection_configuration_url,
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token']),
                                                               connection_object).content)
            result = response.get('result', False)
            if not result:
                self.logger.error(f"Failed to configure the source connection for {source_id} ")
                print(f"Failed to configure the source connection for {source_id}")
                return SourceResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                   error_desc=response, job_id=None, source_id=None)
            else:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS,response=response)
        except Exception as e:
            raise SourceError(f"Failed to configure the source connection for {source_id} " + str(e))

    def source_test_connection_job_poll(self, source_id, poll_timeout=300, polling_frequency=15, retries=3):
        """
        Function to do test connection
        :param source_id: source identifier entity id
        :type source_id: String
        :param poll_timeout: test connection job timeout in seconds
        :type poll_timeout: integer. Default 300 seconds
        :param polling_frequency: polling frequency for the job in seconds. Default 15 seconds
        :type polling_frequency: integer
        :param retries: Number of retries during job poll. Default is 3
        :type retries: integer
        :return: response dict
        """
        url_for_test_connection = url_builder.get_test_connection_url(self.client_config, source_id)
        test_connection_dict = {
            "job_type": "source_test_connection"
        }
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("POST", url_for_test_connection,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              test_connection_dict).content)
            result = response.get('result', {})
        except Exception as e:
            raise SourceError(f"Failed to create test connection job for {source_id} " + str(e))
        if len(result) == 0 or "id" not in result:
            self.logger.error(f"Failed to create test connection job")
            return SourceResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.GENERIC_ERROR,
                                               error_desc=response, job_id=None, source_id=source_id)
        else:
            job_id = result["id"]
            self.logger.info(response.get("message", ""))
            return self.poll_job(source_id=source_id, job_id=job_id, poll_timeout=poll_timeout,
                                 polling_frequency=polling_frequency,
                                 retries=retries)

    def source_metacrawl_job_poll(self, source_id, poll_timeout=300, polling_frequency=15, retries=3):
        """
        Function to do meta crawl connection
        :param source_id: source identifier entity id
        :type source_id: String
        :param poll_timeout: meta crawl job timeout in seconds
        :type poll_timeout: integer. Default 300 seconds
        :param polling_frequency: polling frequency for the job in seconds. Default 15 seconds
        :type polling_frequency: integer
        :param retries: Number of retries during job poll. Default is 3
        :type retries: integer
        :return: response dict
        """
        url_for_metacrawl = url_builder.get_test_connection_url(self.client_config, source_id)
        test_connection_dict = {
            "job_type": "source_fetch_metadata"
        }
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("POST", url_for_metacrawl,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              test_connection_dict).content)
            result = response.get('result', {})
        except Exception as e:
            raise SourceError(f"Failed to create meta crawl job for {source_id} " + str(e))
        if len(result) == 0 or "id" not in result:
            self.logger.error(f"Failed to create meta crawl job")
            return SourceResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.GENERIC_ERROR,
                                               error_desc=response, job_id=None, source_id=source_id)
        else:
            job_id = result["id"]
            self.logger.info(response.get("message", ""))
            return self.poll_job(source_id=source_id, job_id=job_id, poll_timeout=poll_timeout,
                                 polling_frequency=polling_frequency,
                                 retries=retries)

    def browse_source_tables(self, source_id, filter_tables_properties=None, poll_timeout=300, polling_frequency=15,
                             retries=3, poll=True):
        """
        Function to browse the source based on the filter_tables_properties passed
        :param source_id: source identifier entity id
        :type source_id: String
        :param filter_tables_properties:
        :type filter_tables_properties: json object

        filter_tables_properties = {
            "schemas_filter" : "%dbo",
            "catalogs_filter" : "%",
            "tables_filter" : "%csv_incremental_test",
            "is_data_sync_with_filter" : true,
            "is_filter_enabled" : true
        }

        :param poll_timeout: test connection job timeout in seconds
        :type poll_timeout: integer. Default 300 seconds
        :param polling_frequency: polling frequency for the job in seconds. Default 15 seconds
        :type polling_frequency: integer
        :param retries: Number of retries during job poll. Default is 3
        :type retries: integer
        :param poll: Poll job until its completion
        :type poll: Boolean
        :return: response dict
        """
        try:
            url_for_browse_source = url_builder.browse_source_tables_url(self.client_config, source_id)
            if filter_tables_properties is not None:
                filter_condition = f"?is_filter_enabled=true&tables_filter={filter_tables_properties['tables_filter']}&catalogs_filter={filter_tables_properties['catalogs_filter']}&schemas_filter={filter_tables_properties['schemas_filter']}"
                url_for_browse_source = url_for_browse_source + filter_condition
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_for_browse_source,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            result = response.get('result', {})
        except Exception as e:
            raise SourceError(f"Failed to create browse table job for {source_id} " + str(e))
        if len(result) == 0 and "id" not in result:
            self.logger.error(f"Failed to create browse table job")
            return SourceResponse.parse_result(status=Response.Status.FAILED,
                                               error_code=ErrorCode.GENERIC_ERROR,
                                               error_desc=response, job_id=None,
                                               source_id=source_id)
        else:
            job_id = result.get("id")
            self.logger.info(response.get("message", ""))
            if not poll:
                self.logger.info(f"Browse table job for source {source_id} was submitted.")
                return SourceResponse.parse_result(status=Response.Status.SUCCESS)
            job_status = "running"
            failed_count = 0
            timeout = time.time() + poll_timeout
            while True:
                if time.time() > timeout:
                    break
                try:
                    self.logger.info(f"Failed poll job status count: {failed_count}")
                    url_for_interactive_job_poll = url_builder.interactive_job_poll_url(self.client_config, source_id,
                                                                                        job_id)
                    self.logger.info('url to poll interactive job - ' + url_for_interactive_job_poll)
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", url_for_interactive_job_poll,
                                      IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
                    result = response.get('result', {})
                    if len(result) == 0:
                        self.logger.error(f"Failed to poll interactive job {job_id}")
                        job_status = None
                    else:
                        job_status = result["status"]
                    self.logger.info("Browse source job poll status : " + job_status)
                    if job_status in ["completed", "failed", "aborted"]:
                        break
                    if job_status is None:
                        self.logger.info(f"Error occurred during job {job_id} status poll")
                        if failed_count >= retries - 1:
                            return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                               error_code=ErrorCode.GENERIC_ERROR,
                                                               error_desc=response, job_id=job_id,
                                                               source_id=source_id)
                        failed_count = failed_count + 1
                except Exception as e:
                    self.logger.info("Error occurred during job status poll")
                    if failed_count >= retries - 1:
                        raise SourceError(f"Error occurred during job status poll {source_id} " + str(e))
                    failed_count = failed_count + 1
                time.sleep(polling_frequency)
            if job_status == "completed":
                self.logger.info(f"Browse table job for source {source_id} was successful")
                return SourceResponse.parse_result(status=Response.Status.SUCCESS, job_id=job_id)
            else:
                self.logger.error(f"Browse table job for source {source_id} failed with {job_status}")
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=response, job_id=job_id,
                                                   source_id=source_id)

    def add_tables_to_source(self, source_id, tables_to_add_config, poll_timeout=300, polling_frequency=15,
                             retries=3, poll=True):
        """
        Function to add tables to source
        :param source_id: source identifier entity id
        :type source_id: String
        :param tables_to_add_config: Array of JSON configuration object
        :type tables_to_add_config: List

        tables_to_add_config = [{
                "table_name": "",
                "schema_name": "",
                "table_type": "TABLE",
                "target_table_name": "",
                "target_schema_name": ""
        }]

        :param poll_timeout: test connection job timeout in seconds
        :type poll_timeout: integer. Default 300 seconds
        :param polling_frequency: polling frequency for the job in seconds. Default 15 seconds
        :type polling_frequency: integer
        :param retries: Number of retries during job poll. Default is 3
        :type retries: integer
        :param poll: Poll job until its completion
        :type poll: Boolean
        :return: response dict
        """
        try:
            url_for_add_tables_to_source = url_builder.add_tables_to_source_url(self.client_config, source_id)
            add_tables_dict = {"tables_to_add": tables_to_add_config}
            response = IWUtils.ejson_deserialize(
                self.call_api("POST", url_for_add_tables_to_source,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              add_tables_dict).content)
            result = response.get('result', {})
            self.logger.debug(response)
            if len(result) != 0:
                self.logger.info(f"Added the below table Ids to the source {source_id}")
                self.logger.info(result["added_tables"])
                self.logger.info(response["message"])
                self.logger.info(f"Triggered metacrawl job for tables. Infoworks JobID {result['job_created']}")
                job_id = result['job_created']
                if not poll:
                    self.logger.info(f"Tables added to source {source_id} and metacrawl job was submitted {job_id}")
                    return SourceResponse.parse_result(status=Response.Status.SUCCESS)
                else:
                    return self.poll_job(source_id=source_id, job_id=job_id, poll_timeout=poll_timeout,
                                         polling_frequency=polling_frequency,
                                         retries=retries)
            else:
                self.logger.error(f"Failed to add the tables to the source {source_id}")
                self.logger.debug(response)
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=response, job_id=None,
                                                   source_id=source_id)
        except Exception as e:
            raise SourceError(f"Failed to add the tables to the source {source_id} " + str(e))

    def configure_tables_and_tablegroups(self, source_id, configuration_obj):
        """
        Function to configure tables and table-groups.
        The configuration_obj should be similar to the json object that is the output of GET source configuration API
        :param source_id: source identifier entity id
        :type source_id: String
        :param configuration_obj: Configurations for the tables and tablegroups
        :type configuration_obj: JSON object
        :return:  response dict
        """
        try:
            configure_tables_tg_url = url_builder.configure_tables_and_tablegroups_url(self.client_config, source_id)
            errors = {}
            response = IWUtils.ejson_deserialize(
                self.call_api("POST", configure_tables_tg_url,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              {"configuration": configuration_obj}).content)
            result = response.get('result', {})
            count = 0
            # self.logger.debug(response)
            if len(result) != 0:
                for config_item in result:
                    table_upsert_status = config_item.get('table_upsert_status', None)
                    if table_upsert_status is not None and len(table_upsert_status.get("error", [])) != 0:
                        self.logger.info(f"{table_upsert_status}")
                        count = count + 1
                        errors[count] = table_upsert_status["error"]
                        self.logger.error("Found errors during table and table group configurations")
                        self.logger.error(table_upsert_status.get("error", []))
                    elif table_upsert_status is None:
                        self.logger.info(f"Could not get table insert status from {config_item}")
                    else:
                        pass
                    self.logger.debug(str(config_item))
                if len(errors) != 0:
                    self.logger.error(f"Failed due to multiple errors\n{errors}")
                    return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                       error_code=ErrorCode.GENERIC_ERROR,
                                                       error_desc=f"Failed to configure tables and table groups {response} ",
                                                       job_id=None,
                                                       source_id=source_id)
                else:
                    self.logger.info(f"Successfully configured tables and table groups for the source")
                    return SourceResponse.parse_result(status=Response.Status.SUCCESS)
            else:
                self.logger.error("Failed to configure tables and table groups")
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=f"Failed to configure tables and table groups {response} ",
                                                   job_id=None,
                                                   source_id=source_id)
        except Exception as e:
            traceback.print_exc()
            raise SourceError(f"Failed to configure tables and table groups" + str(e))

    def create_table_group(self, source_id, table_group_obj):
        """
        Function to create table group
        :param source_id: entity identifier for source
        :param table_group_obj:

        table_group_obj = {
         "environment_compute_template": {"environment_compute_template_id": "536592c8ceb69bbbe730d452"},
         "name": "tg_name",
         "max_connections": 1,
         "max_parallel_entities": 1,
         "tables": [
          {"table_id":"1123","connection_quota":100}
         ]
        }

        :return:  response dict
        """
        try:
            create_tg_url = url_builder.create_table_group_url(self.client_config, source_id)
            response = IWUtils.ejson_deserialize(
                self.call_api("POST", create_tg_url,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              table_group_obj).content)
            if len(response.get("result", {})) > 0:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS,
                                                   response=response.get("result")["id"])
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=response, job_id=None,
                                                   source_id=source_id)
        except Exception as e:
            raise SourceError(f"Failed to create table group" + str(e))

    def update_table_group(self, source_id, table_group_id, table_group_obj):
        """
        Function to create table group
        :param source_id: entity identifier for source
        :param table_group_id: entity identifier for tablegroup
        :param table_group_obj:

        table_group_obj = {
         "environment_compute_template": {"environment_compute_template_id": "536592c8ceb69bbbe730d452"},
         "name": "tg_name",
         "max_connections": 1,
         "max_parallel_entities": 1,
         "tables": [
          {"table_id":"1123","connection_quota":100}
         ]
        }
        :return: response dict
        """
        try:
            create_tg_url = url_builder.create_table_group_url(self.client_config, source_id) + f"/{table_group_id}"
            response = IWUtils.ejson_deserialize(
                self.call_api("PUT", create_tg_url,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              table_group_obj).content)
            if len(response.get("result", {})) > 0:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS,
                                                   response=response.get("result")["id"])
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=response, job_id=None,
                                                   source_id=source_id)
        except Exception as e:
            raise SourceError(f"Failed to update table group" + str(e))

    def get_list_of_table_groups(self, source_id, params=None, tg_id=None):
        """
        Function to list the table groups under source
        :param source_id: entity identifier for source
        :param tg_id: id of table group config to fetch
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}

        if tg_id is None:
            url_to_list_tg = url_builder.create_table_group_url(
                self.client_config, source_id) + IWUtils.get_query_params_string_from_dict(params=params)
        else:
            url_to_list_tg = url_builder.create_table_group_url(
                self.client_config, source_id) + f"/{tg_id}"
        tg_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_tg,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                if tg_id is not None:
                    tg_list.extend([result])
                else:
                    while len(result) > 0:
                        tg_list.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
            return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=tg_list)
        except Exception as e:
            self.logger.error("Error in listing table groups")
            raise SourceError("Error in listing table groups" + str(e))

    def delete_table_group(self, source_id, tg_id):
        """
        Function to delete table group under the source
        :param source_id: Entity identifier for source
        :type source_id: String
        :param tg_id: id of table group config to delete
        :type tg_id: String
        :return: response dict
        """
        url_to_delete_tg = url_builder.create_table_group_url(self.client_config, source_id).strip("/") + f"/{tg_id}"
        try:
            response = self.call_api("DELETE",
                                     url_to_delete_tg,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content
            )
            if response.status_code == 200:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS)
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=parsed_response.get("details",
                                                                                  "Error in deleting table group")
                                                   )
        except Exception as e:
            self.logger.error("Error in deleting the table group")
            raise SourceError("Error in deleting table group" + str(e))

    def submit_source_job(self, source_id, body, poll=False, poll_timeout=300, polling_frequency=15,
                          retries=3):
        """
        Function to trigger the jobs related to source
        :param source_id: source entity id
        :param body:

        metadata_job_body_example = {
                "job_type": "source_fetch_metadata",
                "overwrite": true
            }

        ingest_table_body_example = {
            "job_type": "cdc_merge" or "truncate_reload",
            "job_name": "testing",
            "interactive_cluster_id": "536592c8ceb69bbbe730d452",
            "table_ids": [
                     "614618debf711275204a1b1f"
                ]
        }

        ingest_table_group_body_example = {
            "job_type": "truncate_reload",
            "table_group_id": "efb9d3e810c643b9930e1a00"
        }
        :param poll_timeout: test connection job timeout in seconds
        :type poll_timeout: integer. Default 300 seconds
        :param polling_frequency: polling frequency for the job in seconds. Default 15 seconds
        :type polling_frequency: integer
        :param retries: Number of retries during job poll. Default is 3
        :type retries: integer
        :param poll: Poll job until its completion
        :type poll: Boolean
        :return:  response dict
        """
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("POST", url_builder.submit_source_job(self.client_config, source_id),
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              body).content)
            result = response.get('result', {})
            if len(result) != 0:
                job_id = result["id"]
                if not poll:
                    self.logger.info(f"Job successfully submitted for {source_id}. JobID to track is: {job_id}")
                    return SourceResponse.parse_result(status=Response.Status.SUCCESS, job_id=job_id)
                else:
                    return self.poll_job(source_id=source_id, job_id=job_id, poll_timeout=poll_timeout,
                                         polling_frequency=polling_frequency,
                                         retries=retries)
        except Exception as e:
            raise SourceError(f"Failed to create source job: " + str(e))

    def resubmit_source_job(self, job_id, poll=False, poll_timeout=300, polling_frequency=15, retries=3):
        """
        Function to resubmit the jobs
        :param job_id: infoworks ob id of the failed job
        :type job_id: String
        :param poll_timeout: test connection job timeout in seconds
        :type poll_timeout: integer. Default 300 seconds
        :param polling_frequency: polling frequency for the job in seconds. Default 15 seconds
        :type polling_frequency: integer
        :param retries: Number of retries during job poll. Default is 3
        :type retries: integer
        :param poll: Poll job until its completion
        :type poll: Boolean
        :return:  response dict
        """
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("POST", url_builder.resubmit_job_url(self.client_config, job_id),
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            result = response.get('result', {})
            if len(result) != 0:
                job_id = result["id"]
                if not poll:
                    self.logger.info(f"Job resubmitted successfully. JobID to track is: {job_id}")
                    return SourceResponse.parse_result(status=Response.Status.SUCCESS, job_id=job_id)
                else:
                    return self.poll_job(source_id=None, job_id=job_id, poll_timeout=poll_timeout,
                                         polling_frequency=polling_frequency,
                                         retries=retries)
        except Exception as e:
            raise SourceError(f"Failed to create source job: " + str(e))

    def get_list_of_sources(self, params=None):
        """
        Function to list the sources
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_sources = url_builder.list_sources_url(
            self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
        source_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_sources,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                while len(result) > 0:
                    source_list.extend([result])
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", [])
            return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=source_list)
        except Exception as e:
            self.logger.error("Error in listing sources")
            raise SourceError("Error in listing sources " + str(e))

    def update_source(self, source_id, update_body):
        """
        Function to update the source
        :param source_id: Entity identifier for source
        :type source_id: String
        :param update_body: JSON body with key-value pair to update
        :type update_body: JSON dict
        :return: response dict
        """
        try:
            response = self.call_api("PATCH", url_builder.source_info(self.client_config, source_id),
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=update_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content
            )
            if response.status_code == 200:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS)
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=parsed_response.get("details", "Error in updating source")
                                                   )
        except Exception as e:
            self.logger.error("Error in updating the source")
            raise SourceError("Error in updating the source" + str(e))

    def get_source_connection_details(self, source_id):
        """
        Function to get source connection details like jdbc url, source type etc
        :param source_id: Entity identifier for source
        :type source_id: String
        :return: response dict
        """
        try:
            source_connection_configuration_url = url_builder.get_source_connection_details_url(self.client_config,
                                                                                                source_id)
            response = IWUtils.ejson_deserialize(self.call_api("GET", source_connection_configuration_url,
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token']),
                                                               ).content)
            result = response.get('result', False)
            if not result:
                self.logger.error(f"Failed to get the source connection for {source_id} ")
                return SourceResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                   error_desc=response, job_id=None, source_id=None)
            else:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=result)
        except Exception as e:
            raise SourceError(f"Failed to get the source connection for {source_id} " + str(e))

    def add_source_advanced_configuration(self, source_id, config_body):
        """
        Function to add advanced configuration to the source
        :param source_id: Entity identifier for source
        :type source_id: String
        :param config_body: JSON body with key-value advanced config
        :type config_body: JSON dict
        :return: response dict
        """
        try:
            response = self.call_api("POST", url_builder.get_advanced_config_url(self.client_config, source_id),
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content
            )
            if response.status_code == 200:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=parsed_response)
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=parsed_response.get("details",
                                                                                  "Error in adding source advanced configs")
                                                   )
        except Exception as e:
            self.logger.error("Error in adding the source advanced config")
            raise SourceError("Error in adding the source advanced config" + str(e))

    def update_source_advanced_configuration(self, source_id, key, config_body):
        """
        Function to update advanced configuration of the source
        :param source_id: Entity identifier for source
        :type source_id: String
        :param key: Name of advanced config to update
        :type key: String
        :param config_body: JSON body with key-value advanced config
        :type config_body: JSON dict
        :return: response dict
        """
        try:
            response = self.call_api("PUT",
                                     url_builder.get_advanced_config_url(self.client_config, source_id).strip(
                                         "/") + f"/{key}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content
            )
            if response.status_code == 200:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS)
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=parsed_response.get("details",
                                                                                  "Error in updating source advanced configs")
                                                   )
        except Exception as e:
            self.logger.error("Error in updating the source advanced configs")
            raise SourceError("Error in updating the source advanced config" + str(e))

    def get_advanced_configuration_of_sources(self, source_id, params=None, key=None):
        """
        Function to list the advanced configuration of source
        :param source_id: Entity identifier for source
        :type source_id: String
        :param key: Name of advanced config to get
        :type key: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}
        if key is None:
            url_to_list_adv_config = url_builder.get_advanced_config_url(
                self.client_config, source_id) + IWUtils.get_query_params_string_from_dict(params=params)
        else:
            url_to_list_adv_config = url_builder.get_advanced_config_url(
                self.client_config, source_id).strip("/") + f"/{key}"
        adv_config_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_adv_config,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                if key is not None:
                    adv_config_list.extend([result])
                else:
                    while len(result) > 0:
                        adv_config_list.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
            return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=adv_config_list)
        except Exception as e:
            self.logger.error("Error in listing advanced configurations")
            raise SourceError("Error in listing advanced configurations" + str(e))

    def delete_source_advanced_configuration(self, source_id, key):
        """
        Function to delete advanced configuration of the source
        :param source_id: Entity identifier for source
        :type source_id: String
        :param key: Name of advanced config to delete
        :type key: String
        :return: response dict
        """
        try:
            response = self.call_api("DELETE",
                                     url_builder.get_advanced_config_url(self.client_config, source_id).strip(
                                         "/") + f"/{key}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content
            )
            if response.status_code == 200:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS)
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=parsed_response.get("details",
                                                                                  "Error in deleting source advanced config")
                                                   )
        except Exception as e:
            self.logger.error("Error in deleting the source advanced config")
            raise SourceError("Error in deleting advanced configurations" + str(e))

    def get_source_configurations_json_export(self, source_id):
        """
        Function to get source configurations
        :param source_id: Entity identifier for source
        :type source_id: String
        :return: response dict
        """
        try:
            source_configuration_url = url_builder.configure_source_url(self.client_config, source_id)
            response = IWUtils.ejson_deserialize(self.call_api("GET", source_configuration_url,
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token']),
                                                               ).content)
            result = response.get('result', False)
            if not result:
                self.logger.error(f"Failed to get the source configurations for {source_id} ")
                return SourceResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                   error_desc=response, job_id=None, source_id=None)
            else:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=result)
        except Exception as e:
            raise SourceError(f"Failed to get the source configurations for {source_id} " + str(e))

    def post_source_configurations_json_import(self, source_id, config_body):
        """
        Function to configure the source
        :param source_id: Entity identifier for source
        :type source_id: String
        :param config_body: JSON body
        :type config_body: JSON dict
        :return: response dict
        """
        source_configuration_url = url_builder.configure_source_url(self.client_config, source_id)
        try:
            response = self.call_api("POST",
                                     source_configuration_url,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=config_body)
            parsed_response = IWUtils.ejson_deserialize(
                self.call_api("POST",
                              source_configuration_url,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              data=config_body).content
            )
            if response.status_code == 200:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS)
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=parsed_response.get("details",
                                                                                  "Error in updating source configs")
                                                   )
        except Exception as e:
            self.logger.error("Error in updating the source configs")
            raise SourceError("Error in updating the source configs" + str(e))

    def list_tables_in_source(self, source_id, params=None):
        """
        Function to list the tables part of the source
        :param source_id: Entity identifier for source
        :type source_id: String
        :param params: Pass the parameters like limit, filter, offset, sort_by, order_by as a dictionary
        :type: JSON dict
        :return: response dict
        """
        if params is None:
            params = {"limit": 20, "offset": 0}
        url_to_list_tables = url_builder.list_tables_under_source(
            self.client_config, source_id) + IWUtils.get_query_params_string_from_dict(params=params)

        tables_list = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_tables,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                while len(result) > 0:
                    tables_list.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", [])
            return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=tables_list)
        except Exception as e:
            self.logger.error("Error in listing tables under source")
            raise SourceError("Error in listing tables under source" + str(e))

    def get_table_columns_details(self, source_id, table_name, schema_name, database_name):
        """
        Function to get the table column details
        :param source_id: Entity identifier for source
        :type source_id: String
        :param table_name: Snowflake Table Name
        :type table_name: String
        :param schema_name: Snowflake Schema Name
        :type schema_name: String
        :param database_name: Snowflake Database Name
        :type database_name: String
        :return: response dict
        """
        url_to_list_tables = url_builder.list_tables_under_source(self.client_config, source_id)
        filter_cond = "?filter={\"table\":\"" + table_name + "\",\"catalog_name\":\"" + database_name + "\",\"schemaNameAtSource\": \"" + schema_name + "\"} "
        get_source_table_info_url = url_to_list_tables + filter_cond
        try:
            response = self.call_api("GET", get_source_table_info_url,
                                     headers=IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                                     )
            if response.status_code != 200:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=f"Failed to get metasync table information for {table_name}")
            else:
                if IWUtils.ejson_deserialize(response.content).get('message') == "No Tables found for given source":
                    return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                       error_code=ErrorCode.GENERIC_ERROR,
                                                       error_desc=f"No Tables found for given source with name {table_name}")

                result = IWUtils.ejson_deserialize(response.content).get('result', [])
                if len(result) != 0:
                    lookup_columns = result[0]["columns"]
                    lookup_columns_dict = {}
                    for column in lookup_columns:
                        lookup_columns_dict[column["name"]] = {}
                        lookup_columns_dict[column["name"]]["target_sql_type"] = column["target_sql_type"]
                        lookup_columns_dict[column["name"]]["target_precision"] = column.get("target_precision", "")
                        lookup_columns_dict[column["name"]]["target_scale"] = column.get("target_scale", "")
                        lookup_columns_dict[column["name"]]["col_size"] = column.get("col_size", "")
                    return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=lookup_columns_dict)
        except Exception as e:
            self.logger.error(f"Error in getting column details for table {table_name}")
            raise SourceError(f"Error in getting column details for table {table_name}" + str(e))

    def get_table_configurations(self, source_id, table_id, ingestion_config_only=False):
        """
        Function to get table configurations
        :param table_id: Entity identifier for table
        :type table_id: String
        :param source_id: Entity identifier for source
        :type source_id: String
        :param ingestion_config_only: True/False. Config to retrieve only ingestion configurations
        :type ingestion_config_only: Boolean
        :return: response dict
        """
        try:
            table_configurations_url = url_builder.get_table_configuration(self.client_config, source_id, table_id)
            if ingestion_config_only:
                table_configurations_url = table_configurations_url + "/configurations/ingestion"
            response = IWUtils.ejson_deserialize(self.call_api("GET", table_configurations_url,
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token']),
                                                               ).content)
            result = response.get('result', False)
            if not result:
                self.logger.error(f"Failed to get the table configurations for {table_id} ")
                return SourceResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                   error_desc=response, job_id=None, source_id=None)
            else:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=result)
        except Exception as e:
            raise SourceError(f"Failed to get the table configurations for {table_id} " + str(e))

    def update_table_configuration(self, source_id, table_id, config_body):
        """
        Function to update table configuration
        :param source_id: Entity identifier for source
        :type source_id: String
        :param table_id: Entity identifier for table
        :type table_id: String
        :param config_body: JSON body with key-value configs
        :type config_body: JSON dict
        :return: response dict
        """
        try:
            response = self.call_api("PUT",
                                     url_builder.get_table_configuration(self.client_config, source_id, table_id),
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content
            )
            if response.status_code == 200:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS)
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=parsed_response.get("details",
                                                                                  "Error in updating table configuration")
                                                   )
        except Exception as e:
            self.logger.error("Error in updating the table configuration")
            raise SourceError(f"Error in updating the table configuration for {table_id} " + str(e))

    def add_table_advanced_configuration(self, source_id, table_id, config_body):
        """
        Function to add advanced configuration to the source
        :param source_id: Entity identifier for source
        :type source_id: String
        :param table_id: Entity identifier for table
        :type table_id: String
        :param config_body: JSON body with key-value advanced config
        :type config_body: JSON dict
        :return: response dict
        """
        try:
            response = self.call_api("POST",
                                     url_builder.get_advanced_config_url(self.client_config, source_id, table_id),
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content
            )
            if response.status_code == 200:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS)
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=parsed_response.get("details",
                                                                                  "Error in adding table advanced config")
                                                   )
        except Exception as e:
            self.logger.error("Error in adding table level advanced configuration")
            raise SourceError(f"Error in adding table level advanced configuration for {table_id} " + str(e))

    def update_table_advanced_configuration(self, source_id, table_id, key, config_body):
        """
        Function to update advanced configuration of the source
        :param source_id: Entity identifier for source
        :type source_id: String
        :param table_id: Entity identifier for table
        :type table_id: String
        :param key: Name of advanced config to update
        :type key: String
        :param config_body: JSON body with key-value advanced config
        :type config_body: JSON dict
        :return: response dict
        """
        try:
            response = self.call_api("PUT",
                                     url_builder.get_advanced_config_url(self.client_config, source_id, table_id).strip(
                                         "/") + f"/{key}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content
            )
            if response.status_code == 200:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS)
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=parsed_response.get("details",
                                                                                  "Error in updating table advanced config")
                                                   )
        except Exception as e:
            self.logger.error("Error in updating the table advanced configs")
            raise SourceError(f"Error in updating table level advanced configuration for {table_id} " + str(e))

    def delete_table_advanced_configuration(self, source_id, table_id, key):
        """
        Function to delete advanced configuration of the table
        :param source_id: Entity identifier for source
        :type source_id: String
        :param table_id: Entity identifier for table
        :type table_id: String
        :param key: Name of advanced config to delete
        :type key: String
        :return: response dict
        """
        try:
            response = self.call_api("DELETE",
                                     url_builder.get_advanced_config_url(self.client_config, source_id, table_id).strip(
                                         "/") + f"/{key}",
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']))
            parsed_response = IWUtils.ejson_deserialize(
                response.content
            )
            if response.status_code == 200:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS)
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=parsed_response.get("details",
                                                                                  "Error in deleting table advanced config")
                                                   )
        except Exception as e:
            self.logger.error("Error in deleting the table advanced config")
            raise SourceError(f"Error in deleting table level advanced configuration for {table_id} " + str(e))

    def get_table_export_configurations(self, source_id, table_id, connection_only=False):
        """
        Function to get table export configurations
        :param connection_only: Get export configuration connection details only
        :param table_id: Entity identifier for table
        :type table_id: String
        :param source_id: Entity identifier for source
        :type source_id: String
        :return: response dict
        """
        try:
            if connection_only:
                table_export_configurations_url = url_builder.table_export_config_url(self.client_config, source_id,
                                                                                      table_id) + "/connection"
            else:
                table_export_configurations_url = url_builder.table_export_config_url(self.client_config, source_id,
                                                                                      table_id)
            response = IWUtils.ejson_deserialize(self.call_api("GET", table_export_configurations_url,
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token']),
                                                               ).content)
            result = response.get('result', False)
            if not result:
                self.logger.error(f"Failed to get the table export configurations for {table_id} ")
                return SourceResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                   error_desc=response, job_id=None, source_id=None)
            else:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=result)
        except Exception as e:
            raise SourceError(f"Failed to get the table export configurations for {table_id} " + str(e))

    def update_table_export_configuration(self, source_id, table_id, config_body, connection_only=False):
        """
        Function to update export configuration of the table
        :param source_id: Entity identifier for source
        :type source_id: String
        :param table_id: Entity identifier for table
        :type table_id: String
        :param config_body: JSON body with key-value pairs
        :type config_body: JSON dict
        :param connection_only: Get export configuration connection details only
        :return: response dict
        """
        if connection_only:
            table_export_configurations_url = url_builder.table_export_config_url(self.client_config, source_id,
                                                                                  table_id) + "/connection"
        else:
            table_export_configurations_url = url_builder.table_export_config_url(self.client_config, source_id,
                                                                                  table_id)
        try:
            response = self.call_api("PUT",
                                     table_export_configurations_url,
                                     IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                     data=config_body)
            parsed_response = IWUtils.ejson_deserialize(
                response.content
            )
            if response.status_code == 200:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS)
            else:
                return SourceResponse.parse_result(status=Response.Status.FAILED,
                                                   error_code=ErrorCode.GENERIC_ERROR,
                                                   error_desc=parsed_response.get("details",
                                                                                  "Error in updating export configuration of the table")
                                                   )
        except Exception as e:
            self.logger.error("Error in updating the export configuration of the table")
            raise SourceError(f"Error in updating the export configuration of the table for {table_id} " + str(e))

    def get_table_ingestion_metrics(self, source_id, table_id):
        """
        Function to fetch ingestion metrics of source tables
        :param source_id: Entity identifier for source
        :type source_id: String
        :param table_id: table entity id
        :type table_id: String
        :return: response dict
        """
        if table_id is None:
            url_to_get_ing_metrics = url_builder.get_ingestion_metrics_source_url(self.client_config, source_id)
        else:
            url_to_get_ing_metrics = url_builder.get_ingestion_metrics_table_url(self.client_config, source_id,
                                                                                 table_id)
        metric_results = []
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_ing_metrics,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                while len(result) > 0:
                    metric_results.extend(result)
                    nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                      ip=self.client_config['ip'],
                                                                      port=self.client_config['port'],
                                                                      protocol=self.client_config['protocol'],
                                                                      )
                    response = IWUtils.ejson_deserialize(
                        self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                            self.client_config['bearer_token'])).content)
                    result = response.get("result", [])
            return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=metric_results)
        except Exception as e:
            self.logger.error("Error in fetching ingestion metrics for source")
            raise SourceError(f"Error in fetching ingestion metrics for source for {source_id} " + str(e))

    def add_table_and_file_mappings_for_csv(self, source_id, file_mappings_config):
        """
        Function to Create table and Add a new file mapping (table) in the csv source
        :param source_id: Entity identifier for source
        :type source_id: String
        :file_mappings_config : Configurations for file_mappings
        :type file_mappings_config: JSON object

        file_mappings_config example = {
                "configuration": {
                    "source_file_properties": {
                        "column_enclosed_by": "\"",
                        "column_separator": ",",
                        "encoding": "UTF-8",
                        "escape_character": "\\",
                        "header_rows_count": 1
                    },
                    "target_relative_path": "/table_path",
                    "deltaLake_table_name": "table_name",
                    "source_file_type": "csv",
                    "ingest_subdirectories": true,
                    "source_relative_path": "",
                    "exclude_filename_regex": "",
                    "include_filename_regex": "",
                    "is_archive_enabled": false,
                    "target_schema_name": "target_schema_name",
                    "target_table_name": "target_table_name"
                },
                "source": "src_id",
                "name": "table_name"
            }
        :return: response dict
        """
        try:
            add_filemappings_url = url_builder.list_tables_under_source(self.client_config, source_id)
            response = IWUtils.ejson_deserialize(self.call_api("POST", add_filemappings_url,
                                                               IWUtils.get_default_header_for_v3(
                                                                   self.client_config['bearer_token']),
                                                               file_mappings_config).content)
            result = response.get('result', False)
            if not result:
                self.logger.error(f"Failed to add table and configure file mappings ")
                return SourceResponse.parse_result(status=Response.Status.FAILED, error_code=ErrorCode.USER_ERROR,
                                                   error_desc=response, job_id=None, source_id=None)
            else:
                return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=result)
        except Exception as e:
            raise SourceError(f"Failed to add table and configure file mappings " + str(e))

    def get_sourceid_from_name(self, source_name):
        """
        Function to return source id from name
        :param source_name: Source name
        :return: Source Identifier
        """
        params = {"filter": {"name": source_name}}
        url_to_list_sources = url_builder.list_sources_url(
            self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_sources,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                if len(result) > 0:
                    return result[0]["id"]
        except Exception as e:
            self.logger.error("Error in listing sources")
            raise SourceError("Error in listing sources " + str(e))

    def get_sourcename_from_id(self, src_id):
        """
        Function to return source name from Id
        :param src_id: Source identifier
        :return: Source name
        """
        url_to_list_sources = url_builder.list_sources_url(
            self.client_config) + f"/{src_id}"
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_sources,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", {})
                return result["name"]
        except Exception as e:
            self.logger.error("Error in getting source name")
            raise SourceError("Error in getting source name " + str(e))

    def create_onboard_rdbms_source(self, src_name, src_type, data_lake_path, connection_url, username, password,
                                    environment_id=None, storage_id=None, schemas_filter=None, tables_filter=None):
        """
        Function to jumpstart creation and onboard of RDBMS source
        :param src_name: Name with which rdbms source has to be created. ["oracle", "mysql", "sqlserver", "teradata", "netezza"]
        :param src_type: Type of the source.
        :param data_lake_path: Path in which data has to be written in datalake
        :param connection_url: JDBC url
        :param username: username for authentication
        :param password: password for authentication. Can be ecrypted using Infoworks encryption algo
        :param environment_id: Entity id of the environment in which the source has to be created
        :param storage_id: Entity id of the source in which the data has to be written
        :param schemas_filter: Filter the schemas to browse
        :param tables_filter: Filter the tables to browse
        """
        if src_type.lower() not in ["oracle", "mysql", "sqlserver", "teradata", "netezza"]:
            print("Invalid source type. Exiting!!!")
            return None
        if src_type.lower() == "oracle":
            sub_type = SourceMappings.oracle.get("sub_type")
            driver_name = SourceMappings.oracle.get("driver_name")
            driver_version = SourceMappings.oracle.get("driver_version")
            database = SourceMappings.oracle.get("database")
        elif src_type.lower() == "sqlserver":
            pass
        elif src_type.lower() == "teradata":
            sub_type = SourceMappings.teradata.get("sub_type")
            driver_name = SourceMappings.teradata.get("driver_name")
            driver_version = SourceMappings.teradata.get("driver_version")
            database = SourceMappings.teradata.get("database")
        elif src_type.lower() == "mysql":
            pass
        else:
            sub_type = SourceMappings.oracle.get("sub_type")
            driver_name = SourceMappings.oracle.get("driver_name")
            driver_version = SourceMappings.oracle.get("driver_version")
            database = SourceMappings.oracle.get("database")
        src_create_response = self.create_source(source_config={
            "name": src_name,
            "type": "rdbms",
            "sub_type": sub_type,
            "data_lake_path": data_lake_path,
            "environment_id": environment_id if environment_id is not None else self.client_config[
                'default_environment_id'],
            "storage_id": storage_id if storage_id is not None else self.client_config['default_storage_id'],
            "is_source_ingested": True
        })
        if src_create_response["result"]["status"].upper() == "SUCCESS":
            src_id = src_create_response["result"]["source_id"]
            response = self.configure_source_connection(src_id, {
                "driver_name": driver_name,
                "driver_version": driver_version,
                "connection_url": connection_url,
                "username": username,
                "password": password,
                "connection_mode": "jdbc",
                "database": database
            })
            if response["result"]["status"].upper() == "SUCCESS":
                response = self.source_test_connection_job_poll(src_id)
                if response["result"]["status"].upper() == "SUCCESS" and response["result"]["job_id"] not in ["", None]:
                    response = self.browse_source_tables(src_id,
                                                         filter_tables_properties={
                                                             "schemas_filter": schemas_filter if schemas_filter is not None else "%",
                                                             "catalogs_filter": "%",
                                                             "tables_filter": tables_filter if tables_filter is not None else "%",
                                                             "is_data_sync_with_filter": True,
                                                             "is_filter_enabled": True
                                                         }, poll_timeout=300, polling_frequency=15, retries=1)
                    if response["result"]["status"].upper() == "SUCCESS" and response["result"]["job_id"] not in ["",
                                                                                                                  None]:
                        print(
                            "Source has been browsed successfully. Please proceed to add tables to source (add_tables_to_source) ")
                        return src_id

    def configure_table_ingestion_properties(self, source_id, table_id, natural_keys=None, sync_type="full-load",
                                             update_strategy=None,
                                             watermark_column=None, partition_key=None, derived_partition=False,
                                             derive_partition_format=None, split_by_key=None, storage_format=None):
        """
        Function to configure the ingestion configs of table
        :param source_id: Entity identifier for source
        :type source_id: String
        :param table_id: table entity id
        :type table_id: String
        :param natural_keys: natural keys List
        :type natural_keys: List
        :param sync_type: full-load/incremental
        :type sync_type: String
        :param update_strategy: append/merge
        :type update_strategy: String
        :param watermark_column: Watermark column to be used during incremental ingestion
        :type watermark_column: String
        :param partition_key: Partition by Column
        :type partition_key: String
        :param derived_partition: True/False. If True pass the derive_partition_format
        :type derived_partition: Boolean
        :param derive_partition_format: Derived partition format. Choose one from ['dd', 'MM', 'yyyy', 'yyyyMM', 'MMdd', 'yyyyMMdd']
        :type derive_partition_format: String
        :param split_by_key: Split by Column
        :type split_by_key: String
        :param storage_format: Storage format for table. orc, parquet, delta
        :type storage_format: String
        :return: response dict
        """
        if natural_keys is None:
            natural_keys = []
        try:
            table_config_body = {}
            if table_id is None:
                self.logger.error("Table id for this method cannot be None")
            if source_id is None:
                self.logger.error("Source id for this method cannot be None")
            if natural_keys != []:
                table_config_body['natural_keys'] = natural_keys
            if sync_type in ['full-load', 'incremental']:
                table_config_body["sync_type"] = "full-load"
                if update_strategy in ["append", "merge"] and watermark_column is not None:
                    if update_strategy == "append":
                        table_config_body["sync_type"] = "incremental"
                        table_config_body["update_strategy"] = "append"
                        table_config_body["watermark_column"] = watermark_column
                    elif update_strategy == "merge" and watermark_column is not None:
                        table_config_body["sync_type"] = "incremental"
                        table_config_body["update_strategy"] = "merge"
                        table_config_body["watermark_column"] = watermark_column
            if partition_key:
                if not derived_partition:
                    table_config_body["partition_keys"] = [{
                        "partition_column": partition_key,
                        "is_derived_partition": False}]
                else:
                    allowed_values_for_date_partition = ['dd', 'MM', 'yyyy', 'yyyyMM', 'MMdd', 'yyyyMMdd']
                    if derive_partition_format in allowed_values_for_date_partition:
                        derive_func_map = {'dd': "day num in month", 'MM': "month", 'yyyy': "year",
                                           'yyyyMM': "year month",
                                           'MMdd': "month day", 'yyyyMMdd': "year month day"}
                        table_config_body["partition_keys"] = [
                            {
                                "parent_column": partition_key,
                                "derive_format": derive_partition_format,
                                "derive_function": derive_func_map[derive_partition_format],
                                "is_derived_partition": True,
                                "partition_column": "iw_partition_column"
                            }]

            if split_by_key:
                table_config_body["split_by_key"] = {
                    "is_derived_split": False,
                    "split_column": split_by_key
                }
            if storage_format is not None:
                table_config_body["storage_format"] = storage_format
            response = IWUtils.ejson_deserialize(
                self.call_api("PUT",
                              url_builder.update_table_ingestion_config_url(self.client_config, source_id, table_id),
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              data=table_config_body).content,
            )
            if response is not None:
                result = response.get("result", [])

            return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error Updating the table ingestion configuration details")
            raise SourceError(
                f"Error Updating the table ingestion configuration details for table for {table_id} " + str(e))

    def configure_table_ingestion_properties_with_payload(self, source_id, table_id, table_payload):
        """
        Function to configure the ingestion configs of table
        :param source_id: Entity identifier for source
        :type source_id: String
        :param table_id: table entity id
        :type table_id: String
        :param table_payload: Payload to configure
        :type table_payload: Dict
        :return: response dict
        """
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("PUT",
                              url_builder.update_table_ingestion_config_url(self.client_config, source_id, table_id),
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                              data=table_payload).content,
            )
            if response is not None:
                result = response.get("result", [])

            return SourceResponse.parse_result(status=Response.Status.SUCCESS, response=response)
        except Exception as e:
            self.logger.error("Error Updating the table ingestion configuration details")
            raise SourceError(
                f"Error Updating the table ingestion configuration details for table for {table_id} " + str(e))

    def get_tableid_from_name(self, source_id, table_name):
        """
                Function to return table id from name
                :param source_id: Source Identifier
                :type source_id: String
                :param table_name: Table name
                :type table_name: String
                :return: Table id
        """
        params = {"filter": {"source": source_id, "table": table_name}}
        url_to_list_tables_under_source = url_builder.list_tables_under_source(
            self.client_config, source_id) + IWUtils.get_query_params_string_from_dict(params=params)
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_tables_under_source,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                if len(result) > 0:
                    return result[0]["id"]
        except Exception as e:
            self.logger.error("Error in listing tables under source")
            raise SourceError("Error in listing tables under source " + str(e))

    def get_tablename_from_id(self, source_id, table_id):
        """
                Function to return table name from id
                :param source_id: Source identifier
                :type source_id: String
                :param table_id: Table identifier
                :type table_id: String
                :return: Table name
        """
        params = {"filter": {"source": source_id, "_id": table_id}}
        url_to_list_tables_under_source = url_builder.list_tables_under_source(
            self.client_config, source_id) + IWUtils.get_query_params_string_from_dict(params=params)
        try:
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_list_tables_under_source,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
            if response is not None:
                result = response.get("result", [])
                if len(result) > 0:
                    return result[0]["name"]
        except Exception as e:
            self.logger.error("Error in listing tables under source")
            raise SourceError("Error in listing tables under source " + str(e))



import traceback

from infoworks.error import WorkflowError
from infoworks.sdk import url_builder
import infoworks.sdk.admin_client as ac
from infoworks.sdk.base_client import BaseClient
from infoworks.sdk.local_configurations import Response
from infoworks.sdk.workflow_response import WorkflowResponse
from infoworks.sdk.utils import IWUtils
import time
from infoworks.sdk.cicd.upload_configurations.workflow import Workflow
from pathlib import Path


class WorkflowClient(BaseClient):
    def __init__(self):
        super(WorkflowClient, self).__init__()

    def get_list_of_workflows(self, domain_id=None, params=None):
        """
        Gets List of Infoworks Data workflow details for given domain id
        :param domain_id: Domain id to which the workflows belongs to, if None all workflows in all domains will be fetched
        :type domain_id: String
        :return: response dict
        """
        response = None
        try:
            if domain_id is None:
                if params is None:
                    params = {"limit": 20, "offset": 0}
                url_to_list_workflows = url_builder.get_all_workflows_url(
                    self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
                workflows_list = []
                response = IWUtils.ejson_deserialize(
                    self.call_api("GET", url_to_list_workflows,
                                  IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
                print(response)
                if response is not None:
                    result = response.get("result", [])
                    while len(result) > 0:
                        workflows_list.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.json().get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
                return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, response=workflows_list)
            else:
                workflows_list = []
                response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.create_workflow_url(
                    self.client_config, domain_id), IWUtils.get_default_header_for_v3(
                    self.client_config['bearer_token'])).content)

                result = response.get('result', {})

                if response is not None:
                    result = response.get("result", [])
                    while len(result) > 0:
                        workflows_list.extend(result)
                        nextUrl = '{protocol}://{ip}:{port}{next}'.format(next=response.get('links')['next'],
                                                                          ip=self.client_config['ip'],
                                                                          port=self.client_config['port'],
                                                                          protocol=self.client_config['protocol'],
                                                                          )
                        response = IWUtils.ejson_deserialize(
                            self.call_api("GET", nextUrl, IWUtils.get_default_header_for_v3(
                                self.client_config['bearer_token'])).content)
                        result = response.get("result", [])
                return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, response=workflows_list)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get workflow details.')
            raise WorkflowError('Error occurred while trying to get workflow details.')

    def create_workflow(self, domain_id=None, workflow_config=None):
        """
        Create a new Workflow
        :param domain_id: Domain id of the workflow
        :type domain_id:String
        :param workflow_config: a JSON object containing workflow configurations
        :type workflow_config: JSON Object

        workflow_config_example = {
        "name": "workflow_name"
        }
        :return: response dict
        """
        response = None
        try:
            if domain_id is None:
                self.logger.error('Please pass the mandatory domain_id as parameter.')
                raise WorkflowError('Please pass the mandatory domain_id as parameter.')

            if workflow_config is None:
                self.logger.error('Invalid workflow configuration. Cannot create a new workflow.')
                raise WorkflowError('Please pass the valid workflow_config as parameter.')
            response = IWUtils.ejson_deserialize(self.call_api("POST", url_builder.create_workflow_url(
                self.client_config, domain_id), IWUtils.get_default_header_for_v3(self.client_config['bearer_token']),
                                                               workflow_config).content)

            result = response.get('result', {})
            workflow_id = result.get('id', None)

            if workflow_id is None:
                self.logger.error('Workflow failed to create.')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc='Workflow failed to create.', response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Workflow {id} has been created under domain {domain_id}.'.format(id=workflow_id, domain_id=domain_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to create a new workflow.')
            raise WorkflowError('Error occurred while trying to create a new workflow.')

    def get_workflow_details(self, workflow_id=None, domain_id=None):
        """
        Gets Infoworks Data workflow details for given workflow id
        :param workflow_id: id of the workflow whose details are to be fetched
        :type workflow_id: String
        :param domain_id: Domain id to which the workflow belongs to
        :type domain_id: String
        :return: response dict
        """
        response = None
        try:
            if workflow_id is None:
                self.logger.error('Please pass the mandatory workflow id as parameter.')
                raise WorkflowError('Please pass the mandatory workflow_id as parameter.')
            if domain_id is None:
                self.logger.error('Please pass the mandatory domain_id as parameter.')
                raise WorkflowError('Please pass the mandatory domain_id as parameter.')
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.create_workflow_url(
                self.client_config, domain_id) + f"/{workflow_id}", IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            result = response.get('result', {})

            if result.get('id', None) is None:
                self.logger.error('Failed to find the workflow details')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc='Failed to find the workflow details',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully got the workflow {id} details.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get workflow details.')
            raise WorkflowError('Error occurred while trying to get workflow details.')

    def delete_workflow(self, workflow_id=None, domain_id=None):
        """
        Deletes Infoworks Data workflow  for given workflow id
        :param workflow_id: entity id of the workflow to be deleted
        :type workflow_id: String
        :param domain_id: Domain id to which the workflow belongs to
        :type domain_id: String
        :return: response dict
        """
        response = None
        try:
            if workflow_id is None:
                self.logger.error('Please pass the mandatory workflow id as parameter.')
                raise WorkflowError('Please pass the mandatory workflow_id as parameter.')
            if domain_id is None:
                self.logger.error('Please pass the mandatory domain_id as parameter.')
                raise WorkflowError('Please pass the mandatory domain_id as parameter.')
            response = IWUtils.ejson_deserialize(self.call_api("DELETE", url_builder.create_workflow_url(
                self.client_config, domain_id) + f"/{workflow_id}", IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            result = response.get('result', {})

            if result.get('id', None) is None:
                self.logger.error(f'Failed to delete the workflow {workflow_id}')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc=f'Failed to delete the workflow {workflow_id}',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully deleted the workflow {id}.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to delete workflow.')
            raise WorkflowError('Error occurred while trying to delete workflow.')

    def update_workflow(self, workflow_id=None, domain_id=None, workflow_config=None):
        """
        Updates Infoworks Data workflow details for given workflow id
        :param workflow_id: entity id of the workflow to be updated
        :type workflow_id: String
        :param domain_id: Domain id to which the workflow belongs to
        :type domain_id: String
        :param workflow_config: a JSON object containing workflow configurations
        :type workflow_config: JSON Object

        workflow_config_example = {
         "name": "{string}",
         "description": "{string}",
         "child_workflow_ids": [
          "{array[string]...}"
         ],
         "domainId": "{string}",
         "workflow_graph": {
          "tasks": [
           {
            "task_id": "e77850ad5127a2d7dab870ff",
            "task_type": "ingest_table_group",
            "location": "-237 52",
            "title": "ingest",
            "task_properties": {},
            "run_properties": {
             "enable_exponential_backoff_for_retries": true,
             "num_retries": 5,
             "retry_delay": 200,
             "max_retry_delay": 1000,
             "trigger_rule": "{string}"
            },
            "workflow_variables": "{}"
           }
          ],
          "edges": [
           {
            "category": "{string}",
            "from_task": "{string}",
            "to_task": "{string}"
           }
          ]
         }
        }
        :return: response dict
        """
        response = None
        try:
            if workflow_config is None:
                self.logger.error('Please pass the mandatory workflow_config as parameter.')
                raise WorkflowError('Please pass valid workflow_config parameter.')
            if workflow_id is None:
                self.logger.error('Please pass the mandatory workflow id as parameter.')
                raise WorkflowError('Please pass the mandatory workflow_id as parameter.')
            if domain_id is None:
                self.logger.error('Please pass the mandatory domain_id as parameter.')
                raise WorkflowError('Please pass the mandatory domain_id as parameter.')
            if "childWorkflowIds" not in workflow_config:
                workflow_config["child_workflow_ids"] = []
            response = IWUtils.ejson_deserialize(self.call_api("PATCH", url_builder.create_workflow_url(
                self.client_config, domain_id) + f"/{workflow_id}", IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token']), workflow_config).content)

            result = response.get('result', {})

            if result.get('id', None) is None:
                self.logger.error(f'Failed to update the workflow {workflow_id}')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc=f'Failed to update the workflow {workflow_id}',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully updated the workflow {id}.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to update workflow.')
            raise WorkflowError('Error occurred while trying to update workflow.')

    def trigger_workflow(self, workflow_id=None, domain_id=None):
        """
        Triggers Infoworks Data workflow for given workflow id
        :param workflow_id: entity id of the workflow to be triggered
        :type workflow_id: String
        :param domain_id: Domain id to which the workflow belongs to
        :type domain_id: String
        :return: response dict
        """
        response = None
        try:
            if workflow_id is None:
                self.logger.error('Please pass the mandatory workflow id as parameter.')
                raise WorkflowError('Please pass the mandatory workflow_id as parameter.')
            if domain_id is None:
                self.logger.error('Please pass the mandatory domain_id as parameter.')
                raise WorkflowError('Please pass the mandatory domain_id as parameter.')
            response = IWUtils.ejson_deserialize(self.call_api("POST", url_builder.trigger_workflow_url(
                self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            result = response.get('result', {})
            run_id = result.get('id', None)
            if result.get('id', None) is None:
                self.logger.error(f'Failed to trigger the workflow {workflow_id}')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc=f'Failed to trigger the workflow {workflow_id}',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully triggered the workflow {id} with run id {run_id}.'.format(id=workflow_id, run_id=run_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to trigger workflow.')
            raise WorkflowError('Error occurred while trying to trigger workflow.')

    def get_status_of_workflow(self, workflow_run_id=None, workflow_id=None):
        """
        Fetches status of Infoworks Data workflow for given workflow id and run id
        :param workflow_run_id: run id of the workflow running
        :type workflow_run_id: String
        :param workflow_id: entity id of the workflow running
        :type workflow_id: String
        :return: response dict
        """
        response = None
        try:
            if workflow_run_id is None:
                self.logger.error('Please pass the mandatory workflow_run_id as parameter.')
                raise WorkflowError('Please pass the mandatory workflow_run_id as parameter.')
            if workflow_id is None:
                self.logger.error('Please pass the mandatory workflow id as parameter.')
                raise WorkflowError('Please pass the mandatory workflow_id as parameter.')
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.get_workflow_status_url(
                self.client_config, workflow_id, workflow_run_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            result = response.get('result', {})
            run_id = result.get('id', None)
            if result.get('id', None) is None:
                self.logger.error(f'Failed to fetch status of the workflow {workflow_id}')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc=f'Failed to fetch status of the workflow {workflow_id}',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully got status of the workflow {id} with run id {run_id}.'.format(id=workflow_id,
                                                                                            run_id=run_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=result)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to fetch status of workflow.')
            raise WorkflowError('Error occurred while trying to fetch status of workflow.')

    def poll_workflow_run_till_completion(self, workflow_run_id=None, workflow_id=None, poll_interval=30):
        """
        Polls Infoworks Data workflow for given workflow id and run id
        :param workflow_run_id: run id of the workflow running
        :type workflow_run_id: String
        :param workflow_id: entity id of the workflow running
        :type workflow_id: String
        :param poll_interval:interval in seconds between poll(default 30)
        :type poll_interval : Int
        :return: response dict
        """
        response = None
        try:
            if workflow_run_id is None:
                self.logger.error('Please pass the mandatory workflow_run_id as parameter.')
                raise WorkflowError('Please pass the mandatory workflow_run_id as parameter.')
            if workflow_id is None:
                self.logger.error('Please pass the mandatory workflow id as parameter.')
                raise WorkflowError('Please pass the mandatory workflow_id as parameter.')
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.get_workflow_status_url(
                self.client_config, workflow_id, workflow_run_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)
            result = response.get('result', {})
            workflow_status = result['workflow_status']["state"]
            while workflow_status.lower() not in ['success', 'completed', 'failed', 'aborted', 'canceled']:
                response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.get_workflow_status_url(
                    self.client_config, workflow_id, workflow_run_id), IWUtils.get_default_header_for_v3(
                    self.client_config['bearer_token'])).content)
                result = response.get('result', {})
                workflow_status = result['workflow_status']["state"]
                time.sleep(poll_interval)

            run_id = result.get('id', None)
            if result.get('id', None) is None:
                self.logger.error(f'Failed to poll status of the workflow {workflow_id}')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc=f'Failed to poll status of the workflow {workflow_id}',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully polled status of the workflow {id} with run id {run_id}.'.format(id=workflow_id,
                                                                                               run_id=run_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=result)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to poll status of workflow.')
            raise WorkflowError('Error occurred while trying to poll status of workflow.')

    def get_workflow_configuration_json_export(self, workflow_id=None, domain_id=None):
        """
        Get exported config for workflow with workflow_id in domain with domain_id
        :param workflow_id: id of the workflow whose details is to be fetched
        :type workflow_id: String
        :param domain_id: Domain id to which the workflow belongs to
        :type domain_id: String
        :return: response dict
        """
        response = None
        try:
            if workflow_id is None:
                self.logger.error('Please pass the mandatory workflow id as parameter.')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc='Invalid workflow configuration. '
                                                                'Please pass the mandatory workflow id as parameter.')
            if domain_id is None:
                self.logger.error('Please pass the mandatory domain id as parameter.')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc='Invalid worlflow configuration. '
                                                                'Please pass the mandatory domain id as parameter.')
            response = IWUtils.ejson_deserialize(self.call_api("GET", url_builder.configure_workflow_url(
                self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token'])).content)

            result = response.get('result', {})

            if result is {}:
                self.logger.error('Failed to get the configuration json of workflow')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc='Failed to get the configuration json of workflow',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully got the workflow {id} configuration json.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to get workflow configuration json.')
            raise WorkflowError('Error occurred while trying to get workflow configuration json.')

    def update_workflow_configuration_json_import(self, workflow_id=None, domain_id=None, workflow_config=None):
        """
        Import config for workflow with workflow_id in domain with domain_id
        :param workflow_id: id of the workflow whose details are to be fetched
        :type workflow_id: String
        :param domain_id: Domain id to which the workflow belongs to
        :type domain_id: String
        :param workflow_config: configuration json of the workflow
        :type workflow_config: JSON Object
        :return: response dict
        """
        response = None
        try:
            if workflow_config is None:
                self.logger.error('Please pass the mandatory workflow_config as parameter.')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc='Invalid workflow configuration. '
                                                                'Please pass the mandatory workflow_config as parameter.')
            if workflow_id is None:
                self.logger.error('Please pass the mandatory workflow id as parameter.')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc='Invalid workflow configuration. '
                                                                'Please pass the mandatory workflow id as parameter.')
            if domain_id is None:
                self.logger.error('Please pass the mandatory domain id as parameter.')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc='Invalid workflow configuration. '
                                                                'Please pass the mandatory domain id as parameter.')
            response = IWUtils.ejson_deserialize(self.call_api("PUT", url_builder.configure_workflow_url(
                self.client_config, domain_id, workflow_id), IWUtils.get_default_header_for_v3(
                self.client_config['bearer_token']), workflow_config).content)

            result = response.get('result', [])

            if len(result) == 0:
                self.logger.error('Failed to update the configuration json of workflow')
                return WorkflowResponse.parse_result(status=Response.Status.FAILED,
                                                     error_desc='Failed to update the configuration json of workflow',
                                                     response=response)

            workflow_id = str(workflow_id)
            self.logger.info(
                'Successfully updated the workflow {id} configuration json.'.format(id=workflow_id))
            return WorkflowResponse.parse_result(status=Response.Status.SUCCESS, workflow_id=workflow_id,
                                                 response=response)

        except Exception as e:
            self.logger.error('Response from server: ' + str(response))
            self.logger.exception('Error occurred while trying to update workflow configuration json.')
            raise WorkflowError('Error occurred while trying to update workflow configuration json.')


    def get_workflow_id(self, workflow_name, domain_id=None, domain_name=None):
        """
        Function to get workflow id
        :param workflow_name: Name of the workflow
        :param domain_id: Entity identifier of the domain
        :param domain_name: Entity name of the domain
        :return: Workflow ID
        """
        if domain_id is None and domain_name is None:
            raise Exception(f"Either domain name or domain id has to be passed to get the pipeline id")
        if domain_name is not None and domain_id is None:
            # Find the domain_id
            params = {"filter": {"name": domain_name}}
            url_to_list_domains = url_builder.create_domain_url(
                self.client_config) + IWUtils.get_query_params_string_from_dict(params=params)
            try:
                response = IWUtils.ejson_deserialize(
                    self.call_api("GET", url_to_list_domains,
                                  IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])).content)
                if response is not None:
                    result = response.get("result", [])
                    domain_id = result[0]["id"]
            except Exception as e:
                self.logger.error("Error in listing domains")
                raise WorkflowError("Error in listing domains" + str(e))
        try:
            params = {"filter": {"name": workflow_name}}
            url_to_get_wf_info = url_builder.create_workflow_url(self.client_config,
                                                                       domain_id) + IWUtils.get_query_params_string_from_dict(
                params=params)
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_wf_info,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", None)
                return result[0].get('id')
        except:
            raise WorkflowError("Unable to get pipeline NAME")

    def get_workflow_name(self, domain_id, workflow_id):
        try:
            url_to_get_wf_info = url_builder.create_workflow_url(self.client_config, domain_id)+f"/{workflow_id}"
            response = IWUtils.ejson_deserialize(
                self.call_api("GET", url_to_get_wf_info,
                              IWUtils.get_default_header_for_v3(self.client_config['bearer_token'])
                              ).content)
            if response is not None and "result" in response:
                result = response.get("result", None)
                return result.get('name')
        except:
            raise WorkflowError("Unable to get workflow name")


def get_bearer_token_url(config):
    """
    returns URL to get bearer token  using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to get the bearer_token
    """
    request = '{protocol}://{ip}:{port}/v3/security/token/access'.format(ip=config['ip'], port=config['port'],
                                                                         protocol=config['protocol'])
    return request


def validate_bearer_token_url(config):
    """
    returns URL to validate bearer token using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to validate bearer token
    """
    request = '{protocol}://{ip}:{port}/v3/security/token/validate'.format(ip=config['ip'], port=config['port'],
                                                                           protocol=config['protocol'])
    return request


def list_domains_url(config):
    """
    returns URL to list domains using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list domains
    """
    request = '{protocol}://{ip}:{port}/v3/domains'.format(ip=config['ip'], port=config['port'],
                                                           protocol=config['protocol'])
    return request


def list_sources_url(config):
    """
    returns URL to list sources using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list sources
    """
    request = '{protocol}://{ip}:{port}/v3/sources'.format(ip=config['ip'], port=config['port'],
                                                           protocol=config['protocol'])
    return request


def list_tables_under_source(config, source_id):
    """
    returns URL to list all tables in source using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity identifier
    :type source_id: string
    :return: url to list all tables in source
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables'.format(ip=config['ip'], port=config['port'],
                                                                              protocol=config['protocol'],
                                                                              source_id=source_id)
    return request


def get_general_configs_url(config):
    """
    returns URL to create source using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to get admin configurations
    """
    request = '{protocol}://{ip}:{port}/v3/admin/configuration/general'.format(ip=config['ip'], port=config['port'],
                                                                               protocol=config['protocol'])
    return request


def get_source_details_url(config):
    """
    returns URL to list sources using v3 api
    :param config: client configurations
    :type config: dict
    :param filter: filter criteria
    :type: filter: string
    :return: source details matching filter criteria
    """

    request = '{protocol}://{ip}:{port}/v3/sources'.format(ip=config['ip'], port=config['port'],
                                                           protocol=config['protocol'])

    return request


def source_info(config, source_id):
    """
    returns URL to get source info using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source id
    :type: source_id: string
    :return: URL to get source information
    """

    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}'.format(ip=config['ip'], port=config['port'],
                                                                       protocol=config['protocol'], source_id=source_id)

    return request


def create_source_url(config):
    """
    returns URL to create source using v3 api
    :param config: client configurations
    :type config: dict
    :return: url for the source creation
    """
    request = '{protocol}://{ip}:{port}/v3/sources'.format(ip=config['ip'], port=config['port'],
                                                           protocol=config['protocol'])
    return request


def get_source_configurations_url(config, source_id):
    """
    returns URL to get source configuration using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to get source configuration
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/configurations'.format(ip=config['ip'],
                                                                                      port=config['port'],
                                                                                      protocol=config['protocol'],
                                                                                      source_id=source_id)
    return request


def get_source_connection_details_url(config, source_id):
    """
    returns URL to get or update source connection details using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to get/update source connection details
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/configurations/connection'.format(ip=config['ip'],
                                                                                                 port=config['port'],
                                                                                                 protocol=config[
                                                                                                     'protocol'],
                                                                                                 source_id=source_id)
    return request


def interactive_job_poll_url(config, source_id, job_id):
    """
    returns URL to import poll interactive jobs using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :param job_id: job id of interactive job
    :type config: string
    :return: url to poll interactive job
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/interactive-jobs/{interactive_job_id}'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        source_id=source_id,
        interactive_job_id=job_id)

    return request


def browse_source_tables_url(config, source_id):
    """
    returns URL to browse source tables using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to browse source tables details
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/source_tables'.format(ip=config['ip'],
                                                                                     port=config['port'],
                                                                                     protocol=config['protocol'],
                                                                                     source_id=source_id)
    return request


def add_tables_to_source_url(config, source_id):
    """
    returns URL to browse source tables using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to browse source tables details
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/source_tables'.format(ip=config['ip'],
                                                                                            port=config['port'],
                                                                                            protocol=config['protocol'],
                                                                                            source_id=source_id)
    return request


def configure_tables_and_tablegroups_url(config, source_id):
    """
    returns URL to configure source tables using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to configure source tables details
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/config-migration'.format(ip=config['ip'],
                                                                                        port=config['port'],
                                                                                        protocol=config['protocol'],
                                                                                        source_id=source_id)
    return request


def create_table_group_url(config, source_id):
    """
        returns URL to create table groups using v3 api
        :param config: client configurations
        :type config: dict
        :param source_id: source entity id
        :type source_id: string
        :return: url to create table groups using v3 api
        """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/table-groups'.format(ip=config['ip'],
                                                                                    port=config['port'],
                                                                                    protocol=config['protocol'],
                                                                                    source_id=source_id)
    return request


def get_compute_template_details_url(config, environment_id):
    """
    returns URL to get compute template details using v3 api
    :param config: client configurations
    :type config: dict
    :param environment_id: environment id
    :type environment_id: string
    :return: url to get compute template details
    """
    request = '{protocol}://{ip}:{port}/v3/admin/environment/{environment_id}/environment-interactive-clusters'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        environment_id=environment_id)
    return request


def get_test_connection_url(config, source_id):
    """
    returns URL to test rdbms source connection using v3 api
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to test rdbms source connection
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/jobs'.format(ip=config['ip'], port=config['port'],
                                                                            protocol=config['protocol'],
                                                                            source_id=source_id)
    return request


def get_job_status_url(config, job_id=None):
    """
    returns URL for checking job progress

    :param config: client configurations
    :type config: dict
    :param job_id: Identifier for Job
    :type job_id: string
    :return: url for the job status
    """
    if job_id is None:
        request = '{protocol}://{ip}:{port}/v3/admin/jobs'.format(ip=config['ip'],
                                                                  port=config['port'],
                                                                  protocol=config['protocol'])
    else:
        request = '{protocol}://{ip}:{port}/v3/admin/jobs/{job_id}'.format(ip=config['ip'],
                                                                           port=config['port'],
                                                                           job_id=job_id,
                                                                           protocol=config['protocol'])
    return request


def configure_source_url(config, source_id):
    """
    returns URL to configure the source using v3 rest apis
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :return: url to configure the source
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/config-migration'.format(ip=config['ip'],
                                                                                        port=config['port'],
                                                                                        protocol=config['protocol'],
                                                                                        source_id=source_id)
    return request


def create_domain_url(config):
    """
    returns URL for creating domain

    :param config: client configurations
    :type config: dict
    :return: url for the domain creation
    """
    request = '{protocol}://{ip}:{port}/v3/domains'.format(ip=config['ip'],
                                                           port=config['port'],
                                                           protocol=config['protocol'])
    return request


def list_users_url(config):
    """
    returns URL to list users using v3 api
    :param config: client configurations
    :type config: dict
    :return: url to list users
    """
    request = '{protocol}://{ip}:{port}/v3/admin/users'.format(ip=config['ip'], port=config['port'],
                                                               protocol=config['protocol'])
    return request


def add_user_to_domain_url(config, user_id):
    """
    returns URL to add user to domains

    :param config: client configurations
    :type config: dict
    :param user_id: Identifier for user
    :type user_id: string
    :return: url for the add user to domain
    """
    request = '{protocol}://{ip}:{port}/v3/admin/users/{user_id}/accessible-domains'.format(ip=config['ip'],
                                                                                            port=config['port'],
                                                                                            user_id=user_id,
                                                                                            protocol=config['protocol'])
    return request


def add_sources_to_domain_url(config, domain_id):
    """
    returns URL to add sources to domains

    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :return: url to add sources to domains
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/accessible-sources'.format(ip=config['ip'],
                                                                                          port=config['port'],
                                                                                          domain_id=domain_id,
                                                                                          protocol=config['protocol'])
    return request


def create_pipeline_url(config, domain_id):
    """
    returns URL to create pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :return: url for the pipeline creation
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines'.format(ip=config['ip'], port=config['port'],
                                                                                 protocol=config['protocol'],
                                                                                 domain_id=domain_id)
    return request


def get_pipeline_url(config, domain_id, pipeline_id):
    """
    returns URL to get pipeline details using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :return: url get pipeline details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}'.format(ip=config['ip'],
                                                                                               port=config['port'],
                                                                                               protocol=config[
                                                                                                   'protocol'],
                                                                                               domain_id=domain_id,
                                                                                               pipeline_id=pipeline_id)
    return request


def delete_pipeline_url(config, domain_id, pipeline_id):
    """
    returns URL to delete pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :return: url to delete pipeline details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}'.format(ip=config['ip'],
                                                                                               port=config['port'],
                                                                                               protocol=config[
                                                                                                   'protocol'],
                                                                                               domain_id=domain_id,
                                                                                               pipeline_id=pipeline_id)
    return request


def update_pipeline_url(config, domain_id, pipeline_id):
    """
    returns URL to update pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :return: url to update pipeline details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}'.format(ip=config['ip'],
                                                                                               port=config['port'],
                                                                                               protocol=config[
                                                                                                   'protocol'],
                                                                                               domain_id=domain_id,
                                                                                               pipeline_id=pipeline_id)
    return request


def configure_pipeline_url(config, domain_id, pipeline_id):
    """
    returns URL to configure pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :return: url to configure pipeline
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/config-migration'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id)
    return request


def list_pipeline_versions_url(config, domain_id, pipeline_id):
    """
    returns URL to list pipeline versions using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :return: url to list pipeline versions
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id)
    return request


def make_pipeline_version_active_url(config, domain_id, pipeline_id, version_id):
    """
    returns URL to make pipeline version active
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :param version_id: Identifier for pipeline
    :type version_id: string
    :return: url to make pipeline version active
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{version_id}/set-active'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id,
        version_id=version_id)
    return request


def create_workflow_url(config, domain_id):
    """
    returns URL to create workflow using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :return: url for the workflow creation
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows'.format(ip=config['ip'], port=config['port'],
                                                                                 protocol=config['protocol'],
                                                                                 domain_id=domain_id)
    return request


def configure_workflow_url(config, domain_id, workflow_id):
    """
    returns URL to configure workflow using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param workflow_id: Identifier for workflow
    :type workflow_id: string
    :return: url to configure workflow
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows/{workflow_id}/config-migration'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        workflow_id=workflow_id)
    return request


def trigger_workflow_url(config, domain_id, workflow_id):
    """
    returns URL to trigger a workflow using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain
    :type domain_id: string
    :param workflow_id: Identifier for workflow
    :type workflow_id: string
    :return: url to trigger workflow
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/workflows/{workflow_id}/start'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        workflow_id=workflow_id)
    return request


def get_workflow_status_url(config, workflow_id, workflow_run_id):
    """
    returns URL to fetch status of workflow using v3 api
    :param config: client configurations
    :type config: dict
    :param workflow_id: Identifier for workflow id
    :type workflow_id: string
    :param workflow_run_id: Identifier for workflow run id
    :type workflow_run_id: string
    :return: url to fetch status of workflow using v3 api
    """
    request = '{protocol}://{ip}:{port}/v3/admin/workflows/{workflow_id}/runs/{workflow_run_id}'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        workflow_id=workflow_id,
        workflow_run_id=workflow_run_id)
    return request


def get_all_workflows_url(config):
    """
    returns URL to fetch all workflows in Infoworks using v3 api
    :param config: client configurations
    :type config: dict
    :return: url fetch all workflows in Infoworks using v3 api
    """
    request = '{protocol}://{ip}:{port}/v3/admin/workflows'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request


def create_data_connection(config, domain_id):
    """
    returns URL to create data connection using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain_id
    :type domain_id: string
    :return: url for data connection creation
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/data-connections'.format(ip=config['ip'],
                                                                                        port=config['port'],
                                                                                        protocol=config['protocol'],
                                                                                        domain_id=domain_id)
    return request


def get_data_connection(config, domain_id, dataconnection_id):
    """
    returns URL to get data connection details using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain_id
    :type domain_id: string
    :param dataconnection_id: Identifier for data connection
    :type dataconnection_id: string
    :return: url to get data connection details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/data-connections/{dataconnection_id}'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id, dataconnection_id=dataconnection_id)
    return request


def get_parent_entity_url(config):
    """
    returns URL to get parent entity
    :param config: client configurations
    :type config: dict
    :return: url to get parent entity
    """
    request = '{protocol}://{ip}:{port}/v3/admin/entity'.format(ip=config['ip'],
                                                                port=config['port'],
                                                                protocol=config['protocol']
                                                                )
    return request


def get_pipeline_details(config, domain_id, pipeline_id):
    """
        returns URL to get pipeline details using v3 api
        :param config: client configurations
        :type config: dict
        :param domain_id: Identifier for domain_id
        :type domain_id: string
        :param pipeline_id: Identifier for pipeline id
        :type pipeline_id: string
        :return: url to get pipeline details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id)
    return request


def get_environment_details(config, env_id=None):
    """
    returns URL to get environment details using v3 api
    :param config: client configurations
    :type config: dict
    :param env_id: Identifier for environment_id
    :type env_id: string
    :return: url to get environment details
    """
    if env_id is not None:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id)
    else:
        request = '{protocol}://{ip}:{port}/v3/admin/environment'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'])
    return request


def get_environment_storage_details(config, env_id, storage_id=None):
    """
    returns URL to get environment storage details using v3 api
    :param config: client configurations
    :type config: dict
    :param env_id: Identifier for environment_id
    :type env_id: string
    :param storage_id: Identifier for storage_id
    :type storage_id: string
    :return: url to get environment storage details
    """
    if storage_id is None:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/environment-storage'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id)
    else:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/environment-storage/{storage_id}'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id,
            storage_id=storage_id)
    return request


def get_environment_compute_details(config, env_id, compute_id=None):
    """
    returns URL to get environment compute details using v3 api
    :param config: client configurations
    :type config: dict
    :param env_id: Identifier for environment_id
    :type env_id: string
    :param compute_id: Identifier for compute_id
    :type compute_id: string
    :return: url to get environment compute details
    """
    if compute_id is None:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/environment-compute-template'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id)
    else:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/environment-compute-template/{compute_id}'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id,
            compute_id=compute_id)
    return request


def get_environment_interactive_compute_details(config, env_id, compute_id=None):
    """
    returns URL to get environment interactive compute details using v3 api
    :param config: client configurations
    :type config: dict
    :param env_id: Identifier for environment_id
    :type env_id: string
    :param compute_id: Identifier for compute_id
    :type compute_id: string
    :return: url to get environment interactive compute details
    """
    if compute_id is None:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/environment-interactive-clusters'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id)
    else:
        request = '{protocol}://{ip}:{port}/v3/admin/environment/{env_id}/environment-interactive-clusters/{compute_id}'.format(
            ip=config['ip'], port=config['port'],
            protocol=config['protocol'],
            env_id=env_id,
            compute_id=compute_id)
    return request


def submit_source_job(config, source_id):
    """
        returns URL to submit source jobs using v3 api
        :param config: client configurations
        :type config: dict
        :param source_id: Identifier for source
        :type source_id: string
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/jobs'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        source_id=source_id)
    return request


def resubmit_job_url(config, job_id):
    """
        returns URL to submit source jobs using v3 api
        :param config: client configurations
        :type config: dict
        :param job_id: Identifier for job
        :type job_id: string
    """
    request = '{protocol}://{ip}:{port}/v3/admin/jobs/{job_id}/resubmit'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        job_id=job_id)
    return request


def get_advanced_config_url(config, source_id, table_id=None):
    """
        returns URL to list advanced configurations of source using v3 api
        :param config: client configurations
        :type config: dict
        :param source_id: source identifier
        :type: source_id: string
        :param table_id: table identifier
        :type: table_id: string
    """
    if table_id is None:
        request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/configurations/advance'.format(ip=config['ip'],
                                                                                                  port=config['port'],
                                                                                                  protocol=config[
                                                                                                      'protocol'],
                                                                                                  source_id=source_id)
    else:
        request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/configurations/advance'.format(
            ip=config['ip'],
            port=config['port'],
            protocol=config[
                'protocol'],
            source_id=source_id, table_id=table_id)

    return request


def get_pipeline_version_url(config, domain_id, pipeline_id, pipeline_version_id):
    """
    returns URL to get pipeline details using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain_id
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline_id
    :type pipeline_id: string
    :return: url get pipeline details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{pipeline_version_id}'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config[
            'protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id,
        pipeline_version_id=pipeline_version_id)
    return request


def delete_pipeline_version_url(config, domain_id, pipeline_id, pipeline_version_id):
    """
    returns URL to delete pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain_id
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline_id
    :type pipeline_id: string
    :return: url delete pipeline
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{pipeline_version_id}'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config[
            'protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id,
        pipeline_version_id=pipeline_version_id)
    return request


def update_pipeline_version_url(config, domain_id, pipeline_id, pipeline_version_id):
    """
    returns URL to update pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain_id
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline_id
    :type pipeline_id: string
    :param pipeline_version_id: Identifier for pipeline version id
    :type pipeline_version_id: string
    :return: url update pipeline details
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{pipeline_version_id}'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config[
            'protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id,
        pipeline_version_id=pipeline_version_id)
    return request


def pipeline_sql_import_url(config, domain_id, pipeline_id):
    """
    returns URL to import sql into pipeline using v3 api
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain_id
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline_id
    :type pipeline_id: string
    :return: url import sql into pipeline
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/sql-import'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config[
            'protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id)
    return request


def get_table_configuration(config, source_id, table_id):
    """
    returns URL to configure the get the table configurations using v3 rest apis
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :param table_id:
    :return: url to configure the source
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}'.format(ip=config['ip'],
                                                                                         port=config['port'],
                                                                                         protocol=config['protocol'],
                                                                                         source_id=source_id,
                                                                                         table_id=table_id)
    return request


def update_table_ingestion_config_url(config, source_id, table_id):
    """
    returns URL to update the table configurations using v3 rest apis
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :param table_id: table entity id
    :type table_id: string
    :return: url to update the table configurations
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/configurations/ingestion'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        source_id=source_id,
        table_id=table_id)
    return request


def table_export_config_url(config, source_id, table_id):
    """
    returns URL to get/update the table export configurations using v3 rest apis
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :param table_id: table entity id
    :type table_id: string
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/configurations/export'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        source_id=source_id,
        table_id=table_id)
    return request


def get_ingestion_metrics_source_url(config, source_id):
    """
    returns URL to get ingestion metrics using v3 rest apis
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/reports/table-metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        source_id=source_id)
    return request


def get_ingestion_metrics_table_url(config, source_id, table_id):
    """
    returns URL to get ingestion metrics using v3 rest apis
    :param config: client configurations
    :type config: dict
    :param source_id: source entity id
    :type source_id: string
    :param table_id: table entity id
    :type table_id: string
    """
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/reports/ingestion-metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        source_id=source_id,
        table_id=table_id)
    return request


def get_job_metrics_url(config, source_id, job_id):
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/jobs/{job_id}/reports/job-metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        job_id=job_id,
        source_id=source_id
    )
    return request


def get_source_file_paths_url(config, source_id, table_id, job_id):
    request = '{protocol}://{ip}:{port}/v3/sources/{source_id}/tables/{table_id}/jobs/{job_id}/reports/sourceFilesPath'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        table_id=table_id,
        job_id=job_id,
        source_id=source_id
    )
    return request


def get_export_metrics_url(config, job_id):
    request = '{protocol}://{ip}:{port}/v3/admin/jobs/{job_id}/reports/export-metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        job_id=job_id
    )
    return request


def get_pipeline_build_metrics_url(config, job_id):
    request = '{protocol}://{ip}:{port}/v3/admin/jobs/{job_id}/reports/pipeline-metrics'.format(
        ip=config['ip'],
        port=config['port'],
        protocol=config['protocol'],
        job_id=job_id
    )
    return request


def get_pipeline_lineage_url(config, domain_id, pipeline_id, pipeline_version_id, column_name, node_id):
    """
        returns URL to get pipeline lineage information using v3 api
        :param config: client configurations
        :type config: dict
        :param domain_id: Identifier for domain_id
        :type domain_id: string
        :param pipeline_id: Identifier for pipeline id
        :type pipeline_id: string
        :param pipeline_version_id: Identifier for pipeline id
        :type pipeline_version_id: string
        :param column_name: Identifier for column name
        :type column_name: string
        :param node_id: Identifier for node id
        :type node_id: string
        :return: url to get pipeline lineage information
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/versions/{pipeline_version_id}/lineage?port_type=OUTPUT&column_name={column_name}&node_id={node_id}'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id,
        pipeline_version_id=pipeline_version_id,
        column_name=column_name,
        node_id=node_id
    )
    return request


def unlock_locked_entities(config):
    """
    returns URL to unlock locked entities
    :param config: client configurations
    :type config: dict
    """
    request = '{protocol}://{ip}:{port}/v3/admin/unlock-entities'.format(ip=config['ip'],
                                                                         port=config['port'],
                                                                         protocol=config['protocol'])
    return request


def trigger_pipeline_build_url(config, domain_id, pipeline_id):
    """
    returns URL to trigger pipeline
    :param config: client configurations
    :type config: dict
    :param domain_id: Identifier for domain_id
    :type domain_id: string
    :param pipeline_id: Identifier for pipeline
    :type pipeline_id: string
    :return: url to trigger pipeline
    """
    request = '{protocol}://{ip}:{port}/v3/domains/{domain_id}/pipelines/{pipeline_id}/jobs'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'],
        domain_id=domain_id,
        pipeline_id=pipeline_id)
    return request


def create_source_extension_url(config):
    """
    returns URL to create source extension
    :param config: client configurations
    :type config: dict
    :return: url to create source extension
    """
    request = '{protocol}://{ip}:{port}/v3/admin/source-extensions'.format(
        ip=config['ip'], port=config['port'],
        protocol=config['protocol'])
    return request
